import pytest

from jenkins_jobs import errors


def dispatch(exc, *args):
    component_type = "type"  # noqa
    name = "name"

    for value in [component_type, name]:
        # prevent pep8 F841 "Unused Variable"
        pass

    raise exc(*args)


def gen_xml(exc, *args):
    data = {"module": "data"}  # noqa

    raise exc(*args)


def test_no_valid_values():
    # When given no valid values, InvalidAttributeError simply displays a
    # message indicating the invalid value, the component type, the
    # component name, and the attribute name.
    message = "'{0}' is an invalid value for attribute {1}.{2}".format(
        "fnord", "type.name", "fubar"
    )
    with pytest.raises(errors.InvalidAttributeError) as excinfo:
        dispatch(errors.InvalidAttributeError, "fubar", "fnord")
    assert str(excinfo.value) == message


def test_with_valid_values():
    # When given valid values, InvalidAttributeError displays a message
    # indicating the invalid value, the component type, the component name,
    # and the attribute name; additionally, it lists the valid values for
    # the current component type & name.
    valid_values = ["herp", "derp"]
    message = "'{0}' is an invalid value for attribute {1}.{2}".format(
        "fnord", "type.name", "fubar"
    )
    message += "\nValid values include: {0}".format(
        ", ".join("'{0}'".format(value) for value in valid_values)
    )

    with pytest.raises(errors.InvalidAttributeError) as excinfo:
        dispatch(errors.InvalidAttributeError, "fubar", "fnord", valid_values)
    assert str(excinfo.value) == message


def test_with_single_missing_attribute():
    # When passed a single missing attribute, display a message indicating
    #  * the missing attribute
    #  * which component type and component name is missing it.
    missing_attribute = "herp"
    message = "Missing {0} from an instance of '{1}'".format(
        missing_attribute, "type.name"
    )

    with pytest.raises(errors.MissingAttributeError) as excinfo:
        dispatch(errors.MissingAttributeError, missing_attribute)
    assert str(excinfo.value) == message

    with pytest.raises(errors.MissingAttributeError) as excinfo:
        gen_xml(errors.MissingAttributeError, missing_attribute)
    assert str(excinfo.value) == message.replace("type.name", "module")


def test_with_multiple_missing_attributes():
    # When passed multiple missing attributes, display a message indicating
    #  * the missing attributes
    #  * which component type and component name is missing it.
    missing_attribute = ["herp", "derp"]
    message = "One of {0} must be present in '{1}'".format(
        ", ".join("'{0}'".format(value) for value in missing_attribute), "type.name"
    )

    with pytest.raises(errors.MissingAttributeError) as excinfo:
        dispatch(errors.MissingAttributeError, missing_attribute)
    assert str(excinfo.value) == message

    with pytest.raises(errors.MissingAttributeError) as excinfo:
        gen_xml(errors.MissingAttributeError, missing_attribute)
    assert str(excinfo.value) == message.replace("type.name", "module")



from unittest.mock import Mock

import pytest

from jenkins_jobs.config import JJBConfig
from jenkins_jobs.yaml_objects import J2String


cases = [
    ("{{ abc }}", {"abc"}),
    ("{% if cond %} {{ x }} {% else %} {{ y }} {% endif %}", {"cond", "x", "y"}),
    ("{# {{ abc }} #}", {}),
]


@pytest.mark.parametrize("format,expected_used_params", cases)
def test_jinja2_required_params(format, expected_used_params):
    config = JJBConfig()
    loader = Mock(source_dir=None)
    template = J2String(config, loader, pos=None, template_text=format)
    assert template.required_params == set(expected_used_params)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dataclasses import dataclass

from .errors import JenkinsJobsException
from .expander import Expander
from .root_base import RootBase, NonTemplateRootMixin, TemplateRootMixin, Group
from .defaults import split_contents_params, job_contents_keys


@dataclass
class JobBase(RootBase):
    project_type: str
    folder: str

    @classmethod
    def from_dict(cls, config, roots, data, pos):
        keep_descriptions = config.yamlparser["keep_descriptions"]
        d = data.copy()
        name = d.pop_required_loc_string("name")
        id = d.pop_loc_string("id", None)
        description = d.pop_loc_string("description", None)
        defaults = d.pop_loc_string("defaults", "global")
        project_type = d.pop_loc_string("project-type", None)
        folder = d.pop_loc_string("folder", None)
        contents, params = split_contents_params(d, job_contents_keys)
        return cls(
            _defaults=roots.defaults,
            _expander=Expander(config),
            _keep_descriptions=keep_descriptions,
            _id=id,
            name=name,
            pos=pos,
            description=description,
            defaults_name=defaults,
            params=params,
            _contents=contents,
            project_type=project_type,
            folder=folder,
        )

    @property
    def contents(self):
        contents = super().contents
        contents["name"] = self.name
        if self.project_type:
            contents["project-type"] = self.project_type
        if self.folder:
            contents["folder"] = self.folder
        return contents

    def _expand_contents(self, contents, params):
        expanded_contents = super()._expand_contents(contents, params)
        try:
            folder = expanded_contents["folder"]
        except KeyError:
            pass
        else:
            name = expanded_contents["name"]
            expanded_contents["name"] = f"{folder}/{name}"
        return expanded_contents


class Job(JobBase, NonTemplateRootMixin):
    @classmethod
    def add(cls, config, roots, data, pos):
        job = cls.from_dict(config, roots, data, pos)
        roots.assign(roots.jobs, job.id, job, "job")

    def __str__(self):
        return f"job {self.name!r}"


class JobTemplate(JobBase, TemplateRootMixin):
    @classmethod
    def add(cls, config, roots, data, pos):
        template = cls.from_dict(config, roots, data, pos)
        roots.assign(roots.job_templates, template.id, template, "job template")

    def __str__(self):
        return f"job template {self.name!r}"


@dataclass
class JobGroup(Group):
    _jobs: dict
    _job_templates: dict

    @classmethod
    def add(cls, config, roots, data, pos):
        d = data.copy()
        name = d.pop_required_loc_string("name")
        try:
            job_specs = cls._specs_from_list(d.pop("jobs", None))
        except JenkinsJobsException as x:
            raise x.with_context(f"In job {name!r}", pos=pos)
        group = cls(
            name,
            pos,
            job_specs,
            d,
            roots.jobs,
            roots.job_templates,
        )
        roots.assign(roots.job_groups, group.name, group, "job group")

    def __str__(self):
        return f"job group {self.name!r}"

    @property
    def _root_dicts(self):
        return [self._jobs, self._job_templates]



# -*- coding: utf-8 -*-
# Copyright (C) 2015 Joost van der Griendt <joostvdg@gmail.com>
# Copyright (C) 2018 Sorin Sbarnea <ssbarnea@users.noreply.github.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The Multibranch Pipeline project module handles creating Jenkins workflow
projects.
You may specify ``multibranch`` in the ``project-type`` attribute of
the :ref:`Job` definition.

Multibranch Pipeline implementantion in JJB is marked as **experimental**
which means that there is no guarantee that its behavior (or configuration)
will not change, even between minor releases.

Plugins required:

* :jenkins-plugins:`Workflow Plugin <workflow-aggregator>`.
* :jenkins-plugins:`Pipeline Multibranch Defaults Plugin
  <pipeline-multibranch-defaults>` (optional)
* :jenkins-plugins:`Basic Branch Build Strategies Plugin
  <basic-branch-build-strategies>` (optional)

:Job Parameters:

    * **scm** (`list`): The SCM definition.

        * **bitbucket** (`dict`): Refer to
          :func:`~bitbucket_scm <bitbucket_scm>` for documentation.

        * **gerrit** (`dict`): Refer to
          :func:`~gerrit_scm <gerrit_scm>` for documentation.

        * **git** (`dict`): Refer to
          :func:`~git_scm <git_scm>` for documentation.

        * **github** (`dict`): Refer to
          :func:`~github_scm <github_scm>` for documentation.

    * **periodic-folder-trigger** (`str`): How often to scan for new branches
      or pull/change requests. Valid values: 1m, 2m, 5m, 10m, 15m, 20m, 25m,
      30m, 1h, 2h, 4h, 8h, 12h, 1d, 2d, 1w, 2w, 4w. (default none)
    * **prune-dead-branches** (`bool`): If dead branches upon check should
      result in their job being dropped. (default true)
    * **number-to-keep** (`int`): How many builds should be kept.
      (default '-1, all')
    * **days-to-keep** (`int`): For how many days should a build be kept.
      (default '-1, forever')
    * **abort-builds** (`bool`): Abort all pending or ongoing builds for removed
      SCM heads (i.e. deleted branches). (default false)
    * **script-path** (`str`): Path to Jenkinsfile, relative to workspace.
      (default 'Jenkinsfile')
    * **script-id** (`str`): Script id from the global Jenkins script store
      provided by the config-file provider plugin. Mutually exclusive with
      **script-path** option.
    * **sandbox** (`bool`): This option is strongly recommended if the
      Jenkinsfile is using load to evaluate a groovy source file from an
      SCM repository. Usable only with **script-id** option. (default 'false')

Job examples:

.. literalinclude:: /../../tests/multibranch/fixtures/multibranch_defaults_id_mode.yaml

.. literalinclude:: /../../tests/multibranch/fixtures/multibranch_defaults_path_mode.yaml

.. literalinclude:: /../../tests/multibranch/fixtures/multi_scm_full.yaml

"""
import collections
import logging
import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
import six

from jenkins_jobs.modules.scm import git_extensions
from jenkins_jobs.errors import InvalidAttributeError, MissingAttributeError
from jenkins_jobs.errors import Context, JenkinsJobsException
from jenkins_jobs.xml_config import remove_ignorable_whitespace

logger = logging.getLogger(str(__name__))


class WorkflowMultiBranch(jenkins_jobs.modules.base.Base):
    sequence = 0
    multibranch_path = "org.jenkinsci.plugins.workflow.multibranch"
    multibranch_defaults_path = "org.jenkinsci.pipeline.workflow.multibranch"
    jenkins_class = "".join([multibranch_path, ".WorkflowMultiBranchProject"])
    jenkins_factory = {
        "script_path": {
            "class": "".join([multibranch_path, ".WorkflowBranchProjectFactory"])
        },
        "script_id": {
            "class": "".join(
                [
                    multibranch_defaults_path,
                    ".defaults.PipelineBranchDefaultsProjectFactory",
                ]
            ),
            "plugin": "pipeline-multibranch-defaults",
        },
    }

    @staticmethod
    def _factory_opts_check(data):

        sandbox = data.get("sandbox", None)
        script_id = data.get("script-id", None)
        script_path = data.get("script-path", None)

        if script_id and script_path:
            error_msg = "script-id and script-path are mutually exclusive options"
            raise JenkinsJobsException(error_msg)
        elif not script_id and sandbox:
            error_msg = (
                "Sandbox mode is applicable only for multibranch with defaults"
                "project type used with script-id option"
            )
            raise JenkinsJobsException(error_msg)

    def root_xml(self, data):
        xml_parent = XML.Element(self.jenkins_class)
        xml_parent.attrib["plugin"] = "workflow-multibranch"
        XML.SubElement(xml_parent, "properties")

        #########
        # Views #
        #########

        views = XML.SubElement(xml_parent, "views")
        all_view = XML.SubElement(views, "hudson.model.AllView")
        all_view_mapping = [
            ("", "name", "All"),
            ("", "filterExecutors", False),
            ("", "filterQueue", False),
        ]
        helpers.convert_mapping_to_xml(
            all_view, {}, all_view_mapping, fail_required=True
        )

        XML.SubElement(
            all_view, "properties", {"class": "hudson.model.View$PropertyList"}
        )

        XML.SubElement(
            all_view, "owner", {"class": self.jenkins_class, "reference": "../../.."}
        )

        XML.SubElement(
            xml_parent, "viewsTabBar", {"class": "hudson.views.DefaultViewsTabBar"}
        )

        ################
        # Folder Views #
        ################

        folderViews = XML.SubElement(
            xml_parent,
            "folderViews",
            {
                "class": "jenkins.branch.MultiBranchProjectViewHolder",
                "plugin": "branch-api",
            },
        )

        XML.SubElement(
            folderViews, "owner", {"class": self.jenkins_class, "reference": "../.."}
        )

        ##################
        # Health Metrics #
        ##################

        hm = XML.SubElement(xml_parent, "healthMetrics")
        hm_path = "com.cloudbees.hudson.plugins.folder.health" ".WorstChildHealthMetric"
        hm_plugin = XML.SubElement(hm, hm_path, {"plugin": "cloudbees-folder"})
        XML.SubElement(hm_plugin, "nonRecursive").text = "false"

        ########
        # Icon #
        ########

        icon = XML.SubElement(
            xml_parent,
            "icon",
            {
                "class": "jenkins.branch.MetadataActionFolderIcon",
                "plugin": "branch-api",
            },
        )
        XML.SubElement(
            icon, "owner", {"class": self.jenkins_class, "reference": "../.."}
        )

        ########################
        # Orphan Item Strategy #
        ########################

        ois_default_strategy = (
            "com.cloudbees.hudson.plugins."
            "folder.computed.DefaultOrphanedItemStrategy"
        )
        ois = XML.SubElement(
            xml_parent,
            "orphanedItemStrategy",
            {"class": ois_default_strategy, "plugin": "cloudbees-folder"},
        )

        ois_mapping = [
            ("prune-dead-branches", "pruneDeadBranches", True, [True, False]),
            ("days-to-keep", "daysToKeep", -1),
            ("number-to-keep", "numToKeep", -1),
            ("abort-builds", "abortBuilds", False, [True, False]),
        ]
        helpers.convert_mapping_to_xml(ois, data, ois_mapping)

        ###########################
        # Periodic Folder Trigger #
        ###########################

        triggers = XML.SubElement(xml_parent, "triggers")

        # Valid options for the periodic trigger interval.
        pft_map = collections.OrderedDict(
            [
                ("1m", ("* * * * *", "60000")),
                ("2m", ("*/2 * * * *", "120000")),
                ("5m", ("*/5 * * * *", "300000")),
                ("10m", ("H/6 * * * *", "600000")),
                ("15m", ("H/6 * * * *", "900000")),
                ("20m", ("H/3 * * * *", "1200000")),
                ("25m", ("H/3 * * * *", "1500000")),
                ("30m", ("H/2 * * * *", "1800000")),
                ("1h", ("H * * * *", "3600000")),
                ("2h", ("H * * * *", "7200000")),
                ("4h", ("H * * * *", "14400000")),
                ("8h", ("H * * * *", "28800000")),
                ("12h", ("H H * * *", "43200000")),
                ("1d", ("H H * * *", "86400000")),
                ("2d", ("H H * * *", "172800000")),
                ("1w", ("H H * * *", "604800000")),
                ("2w", ("H H * * *", "1209600000")),
                ("4w", ("H H * * *", "2419200000")),
            ]
        )

        pft_val = data.get("periodic-folder-trigger")
        if pft_val:
            if not pft_map.get(pft_val):
                raise InvalidAttributeError(
                    "periodic-folder-trigger", pft_val, pft_map.keys()
                )

            pft_path = (
                "com.cloudbees.hudson.plugins.folder.computed." "PeriodicFolderTrigger"
            )
            pft = XML.SubElement(triggers, pft_path, {"plugin": "cloudbees-folder"})
            XML.SubElement(pft, "spec").text = pft_map[pft_val][0]
            XML.SubElement(pft, "interval").text = pft_map[pft_val][1]

        ###########
        # Sources #
        ###########

        sources = XML.SubElement(
            xml_parent,
            "sources",
            {
                "class": "jenkins.branch.MultiBranchProject$BranchSourceList",
                "plugin": "branch-api",
            },
        )
        sources_data = XML.SubElement(sources, "data")
        XML.SubElement(
            sources, "owner", {"class": self.jenkins_class, "reference": "../.."}
        )

        valid_scm = ["bitbucket", "gerrit", "git", "github"]
        for scm_data in data.get("scm", None):
            for scm in scm_data:
                bs = XML.SubElement(sources_data, "jenkins.branch.BranchSource")

                if scm == "bitbucket":
                    bitbucket_scm(bs, scm_data[scm])

                elif scm == "gerrit":
                    gerrit_scm(bs, scm_data[scm])

                elif scm == "git":
                    git_scm(bs, scm_data[scm])

                elif scm == "github":
                    github_scm(bs, scm_data[scm])

                else:
                    raise InvalidAttributeError("scm", scm_data, valid_scm)

        ###########
        # Factory #
        ###########

        self._factory_opts_check(data)

        if data.get("script-id"):
            mode = "script_id"
            fopts_map = (
                ("script-id", "scriptId", None),
                ("sandbox", "useSandbox", None),
            )
        else:
            mode = "script_path"
            fopts_map = (("script-path", "scriptPath", "Jenkinsfile"),)

        factory = XML.SubElement(xml_parent, "factory", self.jenkins_factory[mode])
        XML.SubElement(
            factory, "owner", {"class": self.jenkins_class, "reference": "../.."}
        )

        # multibranch default

        helpers.convert_mapping_to_xml(factory, data, fopts_map, fail_required=False)

        return xml_parent


class WorkflowMultiBranchDefaults(WorkflowMultiBranch):
    multibranch_path = "org.jenkinsci.plugins.workflow.multibranch"
    multibranch_defaults_path = "org.jenkinsci.plugins.pipeline.multibranch"
    jenkins_class = "".join(
        [multibranch_defaults_path, ".defaults.PipelineMultiBranchDefaultsProject"]
    )
    jenkins_factory = {
        "script_path": {
            "class": "".join([multibranch_path, ".WorkflowBranchProjectFactory"]),
            "plugin": "workflow-multibranch",
        },
        "script_id": {
            "class": "".join(
                [
                    multibranch_defaults_path,
                    ".defaults.PipelineBranchDefaultsProjectFactory",
                ]
            )
        },
    }


def bitbucket_scm(xml_parent, data):
    r"""Configure BitBucket scm

    Requires the :jenkins-plugins:`Bitbucket Branch Source Plugin
    <cloudbees-bitbucket-branch-source>`.

    :arg str credentials-id: The credential to use to scan BitBucket.
        (required)
    :arg str repo-owner: Specify the name of the Bitbucket Team or Bitbucket
        User Account. (required)
    :arg str repo: The BitBucket repo. (required)

    :arg bool discover-tags: Discovers tags on the repository.
        (default false)
    :arg bool lfs: Git LFS pull after checkout.
        (default false)
    :arg str server-url: The address of the bitbucket server. (optional)
    :arg str head-filter-regex: A regular expression for filtering
        discovered source branches. Requires the :jenkins-plugins:`SCM API
        Plugin <scm-api>`.
    :arg list head-pr-filter-behaviors: Definition of Filter Branch PR behaviors.
        Requires the :jenkins-plugins:`SCM Filter Branch PR Plugin
        <scm-filter-branch-pr>`.
        Refer to :func:`~add_filter_branch_pr_behaviors <add_filter_branch_pr_behaviors>`.
    :arg str discover-branch: Discovers branches on the repository.
        Valid options: ex-pr, only-pr, all.
        Value is not specified by default.
    :arg str discover-pr-origin: Discovers pull requests where the origin
        repository is the same as the target repository.
        Valid options: mergeOnly, headOnly, mergeAndHead.
        Value is not specified by default.
    :arg str discover-pr-forks-strategy: Fork strategy. Valid options:
        merge-current, current, both, false. (default 'merge-current')
    :arg str discover-pr-forks-trust: Discovers pull requests where the origin
        repository is a fork of the target repository.
        Valid options: contributors, everyone, permission or nobody.
        (default 'contributors')
    :arg list build-strategies: Provides control over whether to build a branch
        (or branch like things such as change requests and tags) whenever it is
        discovered initially or a change from the previous revision has been
        detected. (optional)
        Refer to :func:`~build_strategies <build_strategies>`.
    :arg dict property-strategies: Provides control over how to build a branch
        (like to disable SCM triggering or to override the pipeline durability)
        (optional)
        Refer to :func:`~property_strategies <property_strategies>`.
    :arg bool local-branch: Check out to matching local branch
        If given, checkout the revision to build as HEAD on this branch.
        If selected, then the branch name is computed from the remote branch
        without the origin. In that case, a remote branch origin/master will
        be checked out to a local branch named master, and a remote branch
        origin/develop/new-feature will be checked out to a local branch
        named develop/newfeature.
        Requires the :jenkins-plugins:`Git Plugin <git>`.
    :arg list(str) refspecs: Which refspecs to look for.
    :arg dict checkout-over-ssh: Checkout repo over ssh.

        * **credentials** ('str'): Credentials to use for
            checkout of the repo over ssh.

    :arg dict filter-by-name-wildcard: Enable filter by name with wildcards.
        Requires the :jenkins-plugins:`SCM API Plugin <scm-api>`.

        * **includes** ('str'): Space-separated list
            of name patterns to consider. You may use * as a wildcard;
            for example: `master release*`
        * **excludes** ('str'): Name patterns to
            ignore even if matched by the includes list.
            For example: `release*`

    :extensions:

        * **clean** (`dict`)
            * **after** (`dict`) - Clean the workspace after checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
            * **before** (`dict`) - Clean the workspace before checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
        * **prune** (`bool`) - Prune remote branches (default false)
        * **shallow-clone** (`bool`) - Perform shallow clone (default false)
        * **sparse-checkout** (dict)
            * **paths** (list) - List of paths to sparse checkout. (optional)
        * **depth** (`int`) - Set shallow clone depth (default 1)
        * **do-not-fetch-tags** (`bool`) - Perform a clone without tags
            (default false)
        * **submodule** (`dict`)
            * **disable** (`bool`) - By disabling support for submodules you
              can still keep using basic git plugin functionality and just have
              Jenkins to ignore submodules completely as if they didn't exist.
            * **recursive** (`bool`) - Retrieve all submodules recursively
              (uses '--recursive' option which requires git>=1.6.5)
            * **tracking** (`bool`) - Retrieve the tip of the configured
              branch in .gitmodules (Uses '\-\-remote' option which requires
              git>=1.8.2)
            * **parent-credentials** (`bool`) - Use credentials from default
              remote of parent repository (default false).
            * **reference-repo** (`str`) - Path of the reference repo to use
              during clone (optional)
            * **timeout** (`int`) - Specify a timeout (in minutes) for
              submodules operations (default 10).
        * **timeout** (`str`) - Timeout for git commands in minutes (optional)
        * **use-author** (`bool`): Use author rather than committer in Jenkin's
            build changeset (default false)
        * **wipe-workspace** (`bool`) - Wipe out workspace before build
            (default true)
        * **lfs-pull** (`bool`) - Call git lfs pull after checkout
            (default false)


    Minimal Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_bitbucket_minimal.yaml

    Full Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_bitbucket_full.yaml
    """
    source = XML.SubElement(
        xml_parent,
        "source",
        {
            "class": "com.cloudbees.jenkins.plugins.bitbucket.BitbucketSCMSource",
            "plugin": "cloudbees-bitbucket-branch-source",
        },
    )
    source_mapping = [
        ("", "id", "-".join(["bb", data.get("repo-owner", ""), data.get("repo", "")])),
        ("repo-owner", "repoOwner", None),
        ("repo", "repository", None),
    ]
    helpers.convert_mapping_to_xml(source, data, source_mapping, fail_required=True)

    mapping_optional = [
        ("credentials-id", "credentialsId", None),
        ("server-url", "serverUrl", None),
    ]
    helpers.convert_mapping_to_xml(source, data, mapping_optional, fail_required=False)

    traits = XML.SubElement(source, "traits")

    if data.get("refspecs"):
        refspec_trait = XML.SubElement(
            traits,
            "jenkins.plugins.git.traits.RefSpecsSCMSourceTrait",
            {"plugin": "git"},
        )
        templates = XML.SubElement(refspec_trait, "templates")
        refspecs = data.get("refspecs")
        for refspec in refspecs:
            e = XML.SubElement(
                templates,
                (
                    "jenkins.plugins.git.traits"
                    ".RefSpecsSCMSourceTrait_-RefSpecTemplate"
                ),
            )
            XML.SubElement(e, "value").text = refspec

    if data.get("discover-tags", False):
        XML.SubElement(
            traits, "com.cloudbees.jenkins.plugins.bitbucket.TagDiscoveryTrait"
        )

    if data.get("lfs", False):
        gitlfspull = XML.SubElement(
            traits, "jenkins.plugins.git.traits.GitLFSPullTrait", {"plugin": "git"}
        )
        XML.SubElement(
            gitlfspull,
            "extension",
            {"class": "hudson.plugins.git.extensions.impl.GitLFSPull"},
        )

    if data.get("head-filter-regex", None):
        rshf = XML.SubElement(traits, "jenkins.scm.impl.trait.RegexSCMHeadFilterTrait")
        XML.SubElement(rshf, "regex").text = data.get("head-filter-regex")

    if data.get("head-pr-filter-behaviors", None):
        add_filter_branch_pr_behaviors(traits, data.get("head-pr-filter-behaviors"))

    if data.get("discover-pr-origin", None):
        dpro = XML.SubElement(
            traits,
            "com.cloudbees.jenkins.plugins.bitbucket"
            ".OriginPullRequestDiscoveryTrait",
        )
        dpro_strategies = {"mergeOnly": "1", "headOnly": "2", "mergeAndHead": "3"}
        dpro_mapping = [("discover-pr-origin", "strategyId", None, dpro_strategies)]
        helpers.convert_mapping_to_xml(dpro, data, dpro_mapping, fail_required=True)

    if data.get("discover-pr-forks-strategy"):
        dprf = XML.SubElement(
            traits,
            "com.cloudbees.jenkins.plugins.bitbucket" ".ForkPullRequestDiscoveryTrait",
        )
        dprf_strategy = {"merge-current": "1", "current": "2", "both": "3"}
        dprf_mapping = [
            ("discover-pr-forks-strategy", "strategyId", "merge-current", dprf_strategy)
        ]
        helpers.convert_mapping_to_xml(dprf, data, dprf_mapping, fail_required=True)

        trust = data.get("discover-pr-forks-trust", "contributors")
        trust_map = {
            "contributors": "".join(
                [
                    "com.cloudbees.jenkins.plugins.bitbucket"
                    ".ForkPullRequestDiscoveryTrait$TrustContributors"
                ]
            ),
            "everyone": "".join(
                [
                    "com.cloudbees.jenkins.plugins.bitbucket"
                    ".ForkPullRequestDiscoveryTrait$TrustEveryone"
                ]
            ),
            "permission": "".join(
                [
                    "com.cloudbees.jenkins.plugins.bitbucket"
                    ".ForkPullRequestDiscoveryTrait$TrustPermission"
                ]
            ),
            "nobody": "".join(
                [
                    "com.cloudbees.jenkins.plugins.bitbucket"
                    ".ForkPullRequestDiscoveryTrait$TrustNobody"
                ]
            ),
        }
        if trust not in trust_map:
            raise InvalidAttributeError(
                "discover-pr-forks-trust", trust, trust_map.keys()
            )
        XML.SubElement(dprf, "trust").attrib["class"] = trust_map[trust]

    if data.get("discover-branch", None):
        dbr = XML.SubElement(
            traits, "com.cloudbees.jenkins.plugins.bitbucket.BranchDiscoveryTrait"
        )
        dbr_strategies = {"ex-pr": "1", "only-pr": "2", "all": "3"}
        dbr_mapping = [("discover-branch", "strategyId", None, dbr_strategies)]
        helpers.convert_mapping_to_xml(dbr, data, dbr_mapping, fail_required=True)

    if data.get("property-strategies", None):
        property_strategies(xml_parent, data)

    if data.get("build-strategies", None):
        build_strategies(xml_parent, data)

    if data.get("local-branch", False):
        lbr = XML.SubElement(
            traits, "jenkins.plugins.git.traits.LocalBranchTrait", {"plugin": "git"}
        )
        lbr_extension = XML.SubElement(
            lbr,
            "extension",
            {"class": "hudson.plugins.git.extensions.impl.LocalBranch"},
        )
        XML.SubElement(lbr_extension, "localBranch").text = "**"

    if data.get("checkout-over-ssh", None):
        cossh = XML.SubElement(
            traits, "com.cloudbees.jenkins.plugins.bitbucket.SSHCheckoutTrait"
        )
        cossh_credentials = [("credentials", "credentialsId", "")]
        helpers.convert_mapping_to_xml(
            cossh, data.get("checkout-over-ssh"), cossh_credentials, fail_required=True
        )

    add_filter_by_name_wildcard_behaviors(traits, data)

    # handle the default git extensions like:
    # - clean
    # - shallow-clone
    # - timeout
    # - do-not-fetch-tags
    # - submodule
    # - prune
    # - wipe-workspace
    # - use-author
    # - lfs-pull
    git_extensions(traits, data)


def gerrit_scm(xml_parent, data):
    r"""Configure Gerrit SCM

    Requires the :jenkins-plugins:`Gerrit Code Review Plugin
    <gerrit-code-review>`.

    :arg str url: The git url. (required)
    :arg str credentials-id: The credential to use to connect to the GIT URL.
    :arg bool ignore-on-push-notifications: If a job should not trigger upon
        push notifications. (default false)
    :arg list(str) refspecs: Which refspecs to look for.
        (default ``['+refs/changes/*:refs/remotes/@{remote}/*',
        '+refs/heads/*:refs/remotes/@{remote}/*']``)
    :arg str includes: Comma-separated list of branches to be included.
        (default '*')
    :arg str excludes: Comma-separated list of branches to be excluded.
        (default '')
    :arg str head-filter-regex: A regular expression for filtering
        discovered source branches. Requires the :jenkins-plugins:`SCM API
        Plugin <scm-api>`.
    :arg list build-strategies: Provides control over whether to build a branch
        (or branch like things such as change requests and tags) whenever it is
        discovered initially or a change from the previous revision has been
        detected. (optional)
        Refer to :func:`~build_strategies <build_strategies>`.
    :arg dict property-strategies: Provides control over how to build a branch
        (like to disable SCM triggering or to override the pipeline durability)
        (optional)
        Refer to :func:`~property_strategies <property_strategies>`.
    :arg dict filter-checks: Enable the filtering by pending checks, allowing to
        discover the changes that need validation only. This feature is using
        the gerrit checks plugin.
        (optional)
        query-operator: Name of the query operator, supported values are:
        'SCHEME' or 'ID'.
        query-string: Value of the query operator.
    :arg dict change-discovery: Configure the query string in 'Discover open changes'.
        The default 'p:<project> status:open -age:24w' will be added prior to the
        query-string specified here.
        (optional)
        query-string: Value of the query operator.

    :extensions:

        * **clean** (`dict`)
            * **after** (`dict`) - Clean the workspace after checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
            * **before** (`dict`) - Clean the workspace before checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
        * **prune** (`bool`) - Prune remote branches (default false)
        * **shallow-clone** (`bool`) - Perform shallow clone (default false)
        * **sparse-checkout** (dict)
            * **paths** (list) - List of paths to sparse checkout. (optional)
        * **depth** (`int`) - Set shallow clone depth (default 1)
        * **do-not-fetch-tags** (`bool`) - Perform a clone without tags
            (default false)
        * **submodule** (`dict`)
            * **disable** (`bool`) - By disabling support for submodules you
              can still keep using basic git plugin functionality and just have
              Jenkins to ignore submodules completely as if they didn't exist.
            * **recursive** (`bool`) - Retrieve all submodules recursively
              (uses '--recursive' option which requires git>=1.6.5)
            * **tracking** (`bool`) - Retrieve the tip of the configured
              branch in .gitmodules (Uses '\-\-remote' option which requires
              git>=1.8.2)
            * **parent-credentials** (`bool`) - Use credentials from default
              remote of parent repository (default false).
            * **reference-repo** (`str`) - Path of the reference repo to use
              during clone (optional)
            * **timeout** (`int`) - Specify a timeout (in minutes) for
              submodules operations (default 10).
        * **timeout** (`str`) - Timeout for git commands in minutes (optional)
        * **use-author** (`bool`): Use author rather than committer in Jenkin's
            build changeset (default false)
        * **wipe-workspace** (`bool`) - Wipe out workspace before build
            (default true)
        * **lfs-pull** (`bool`) - Call git lfs pull after checkout
            (default false)

    Minimal Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_gerrit_minimal.yaml

    Full Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_gerrit_full.yaml
    """
    source = XML.SubElement(
        xml_parent,
        "source",
        {
            "class": "jenkins.plugins.gerrit.GerritSCMSource",
            "plugin": "gerrit-code-review",
        },
    )
    source_mapping = [
        ("", "id", "-".join(["gr", data.get("url", "")])),
        ("url", "remote", None),
        ("credentials-id", "credentialsId", ""),
        ("includes", "includes", "*"),
        ("excludes", "excludes", ""),
        ("ignore-on-push-notifications", "ignoreOnPushNotifications", True),
    ]
    helpers.convert_mapping_to_xml(source, data, source_mapping, fail_required=True)

    source_mapping_optional = [("api-uri", "apiUri", None)]
    helpers.convert_mapping_to_xml(
        source, data, source_mapping_optional, fail_required=False
    )

    # Traits
    traits = XML.SubElement(source, "traits")
    change_discovery_trait = XML.SubElement(
        traits, "jenkins.plugins.gerrit.traits.ChangeDiscoveryTrait"
    )
    change_discovery = data.get("change-discovery", None)
    if change_discovery:
        change_discovery_mapping = [("query-string", "queryString", None)]
        helpers.convert_mapping_to_xml(
            change_discovery_trait,
            change_discovery,
            change_discovery_mapping,
            fail_required=True,
        )

    # Refspec Trait
    refspec_trait = XML.SubElement(
        traits, "jenkins.plugins.git.traits.RefSpecsSCMSourceTrait", {"plugin": "git"}
    )
    templates = XML.SubElement(refspec_trait, "templates")
    refspecs = data.get(
        "refspecs",
        [
            "+refs/changes/*:refs/remotes/@{remote}/*",
            "+refs/heads/*:refs/remotes/@{remote}/*",
        ],
    )
    # convert single string to list
    if isinstance(refspecs, six.string_types):
        refspecs = [refspecs]
    for x in refspecs:
        e = XML.SubElement(
            templates,
            ("jenkins.plugins.git.traits" ".RefSpecsSCMSourceTrait_-RefSpecTemplate"),
        )
        XML.SubElement(e, "value").text = x

    if data.get("property-strategies", None):
        property_strategies(xml_parent, data)

    if data.get("build-strategies", None):
        build_strategies(xml_parent, data)

    if data.get("head-filter-regex", None):
        rshf = XML.SubElement(traits, "jenkins.scm.impl.trait.RegexSCMHeadFilterTrait")
        XML.SubElement(rshf, "regex").text = data.get("head-filter-regex")

    # FilterChecks Trait
    checks = data.get("filter-checks", None)
    if checks:
        checks_trait = XML.SubElement(
            traits, "jenkins.plugins.gerrit.traits.FilterChecksTrait"
        )
        checks_source_mapping = [
            ("query-operator", "queryOperator", None),
            ("query-string", "queryString", None),
        ]
        helpers.convert_mapping_to_xml(
            checks_trait, checks, checks_source_mapping, fail_required=True
        )

    # handle the default git extensions like:
    # - clean
    # - shallow-clone
    # - timeout
    # - do-not-fetch-tags
    # - submodule
    # - prune
    # - wipe-workspace
    # - use-author
    # - lfs-pull
    git_extensions(traits, data)


def git_scm(xml_parent, data):
    r"""Configure Git SCM

    Requires the :jenkins-plugins:`Git Plugin <git>`.

    :arg str url: The git repo url. (required)
    :arg str credentials-id: The credential to use to connect to the GIT repo.
        (default '')

    :arg bool discover-branches: Discovers branches on the repository.
        (default true)
    :arg bool discover-tags: Discovers tags on the repository.
        (default false)
    :arg bool ignore-on-push-notifications: If a job should not trigger upon
        push notifications. (default false)
    :arg str head-filter-regex: A regular expression for filtering
        discovered source branches. Requires the :jenkins-plugins:`SCM API
        Plugin <scm-api>`.
    :arg list head-pr-filter-behaviors: Definition of Filter Branch PR behaviors.
        Requires the :jenkins-plugins:`SCM Filter Branch PR Plugin
        <scm-filter-branch-pr>`.
        Refer to :func:`~add_filter_branch_pr_behaviors <add_filter_branch_pr_behaviors>`.
    :arg list build-strategies: Provides control over whether to build a branch
        (or branch like things such as change requests and tags) whenever it is
        discovered initially or a change from the previous revision has been
        detected. (optional)
        Refer to :func:`~build_strategies <build_strategies>`.
    :arg dict property-strategies: Provides control over how to build a branch
        (like to disable SCM triggering or to override the pipeline durability)
        (optional)
        Refer to :func:`~property_strategies <property_strategies>`.

    :extensions:

        * **clean** (`dict`)
            * **after** (`dict`) - Clean the workspace after checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
            * **before** (`dict`) - Clean the workspace before checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
        * **prune** (`bool`) - Prune remote branches (default false)
        * **shallow-clone** (`bool`) - Perform shallow clone (default false)
        * **sparse-checkout** (dict)
            * **paths** (list) - List of paths to sparse checkout. (optional)
        * **depth** (`int`) - Set shallow clone depth (default 1)
        * **do-not-fetch-tags** (`bool`) - Perform a clone without tags
            (default false)
        * **submodule** (`dict`)
            * **disable** (`bool`) - By disabling support for submodules you
              can still keep using basic git plugin functionality and just have
              Jenkins to ignore submodules completely as if they didn't exist.
            * **recursive** (`bool`) - Retrieve all submodules recursively
              (uses '--recursive' option which requires git>=1.6.5)
            * **tracking** (`bool`) - Retrieve the tip of the configured
              branch in .gitmodules (Uses '\-\-remote' option which requires
              git>=1.8.2)
            * **parent-credentials** (`bool`) - Use credentials from default
              remote of parent repository (default false).
            * **reference-repo** (`str`) - Path of the reference repo to use
              during clone (optional)
            * **timeout** (`int`) - Specify a timeout (in minutes) for
              submodules operations (default 10).
        * **timeout** (`str`) - Timeout for git commands in minutes (optional)
        * **use-author** (`bool`): Use author rather than committer in Jenkin's
            build changeset (default false)
        * **wipe-workspace** (`bool`) - Wipe out workspace before build
            (default true)
        * **lfs-pull** (`bool`) - Call git lfs pull after checkout
            (default false)

    Minimal Example:

    .. literalinclude:: /../../tests/multibranch/fixtures/scm_git_minimal.yaml

    Full Example:

    .. literalinclude:: /../../tests/multibranch/fixtures/scm_git_full.yaml
    """
    source = XML.SubElement(
        xml_parent,
        "source",
        {"class": "jenkins.plugins.git.GitSCMSource", "plugin": "git"},
    )
    source_mapping = [
        ("", "id", "-".join(["gt", data.get("url", "")])),
        ("url", "remote", None),
        ("credentials-id", "credentialsId", ""),
    ]
    helpers.convert_mapping_to_xml(source, data, source_mapping, fail_required=True)

    ##########
    # Traits #
    ##########

    traits_path = "jenkins.plugins.git.traits"
    traits = XML.SubElement(source, "traits")

    if data.get("discover-branches", True):
        XML.SubElement(traits, "".join([traits_path, ".BranchDiscoveryTrait"]))

    if data.get("discover-tags", False):
        XML.SubElement(traits, "".join([traits_path, ".TagDiscoveryTrait"]))

    if data.get("ignore-on-push-notifications", False):
        XML.SubElement(traits, "".join([traits_path, ".IgnoreOnPushNotificationTrait"]))

    if data.get("head-filter-regex", None):
        rshf = XML.SubElement(traits, "jenkins.scm.impl.trait.RegexSCMHeadFilterTrait")
        XML.SubElement(rshf, "regex").text = data.get("head-filter-regex")

    if data.get("head-pr-filter-behaviors", None):
        add_filter_branch_pr_behaviors(traits, data.get("head-pr-filter-behaviors"))

    if data.get("property-strategies", None):
        property_strategies(xml_parent, data)

    if data.get("build-strategies", None):
        build_strategies(xml_parent, data)

    # handle the default git extensions like:
    # - clean
    # - shallow-clone
    # - timeout
    # - do-not-fetch-tags
    # - submodule
    # - prune
    # - wipe-workspace
    # - use-author
    # - lfs-pull
    git_extensions(traits, data)


def github_scm(xml_parent, data):
    r"""Configure GitHub SCM

    Requires the :jenkins-plugins:`GitHub Branch Source Plugin
    <github-branch-source>`.

    :arg str api-uri: The GitHub API uri for hosted / on-site GitHub. Must
        first be configured in Global Configuration. (default GitHub)
    :arg bool ssh-checkout: Checkout over SSH.

        * **credentials** ('str'): Credentials to use for
            checkout of the repo over ssh.

    :arg str credentials-id: Credentials used to scan branches and pull
        requests, check out sources and mark commit statuses. (optional)
    :arg str repo-owner: Specify the name of the GitHub Organization or
        GitHub User Account. (required)
    :arg str repo: The GitHub repo. (required)

    :arg str branch-discovery: Discovers branches on the repository.
        Valid options: no-pr, only-pr, all, false. (default 'no-pr')
    :arg str discover-pr-forks-strategy: Fork strategy. Valid options:
        merge-current, current, both, false. (default 'merge-current')
    :arg str discover-pr-forks-trust: Discovers pull requests where the origin
        repository is a fork of the target repository.
        Valid options: contributors, everyone, permission or nobody.
        (default 'contributors')
    :arg str discover-pr-origin: Discovers pull requests where the origin
        repository is the same as the target repository.
        Valid options: merge-current, current, both, false.  (default 'merge-current')
    :arg bool discover-tags: Discovers tags on the repository.
        (default false)
    :arg list head-pr-filter-behaviors: Definition of Filter Branch PR behaviors.
        Requires the :jenkins-plugins:`SCM Filter Branch PR Plugin
        <scm-filter-branch-pr>`.
        Refer to :func:`~add_filter_branch_pr_behaviors <add_filter_branch_pr_behaviors>`.
    :arg list build-strategies: Provides control over whether to build a branch
        (or branch like things such as change requests and tags) whenever it is
        discovered initially or a change from the previous revision has been
        detected. (optional)
        Refer to :func:`~build_strategies <build_strategies>`.
    :arg dict notification-context: Change the default GitHub check notification
        context from "continuous-integration/jenkins/SUFFIX" to a custom label / suffix.
        (set a label and suffix to true or false, optional)
        Requires the :jenkins-plugins:`Github Custom Notification Context SCM
        Behaviour <github-scm-trait-notification-context>`.
        Refer to :func:`~add_notification_context_trait <add_notification_context_trait>`.
    :arg dict property-strategies: Provides control over how to build a branch
        (like to disable SCM triggering or to override the pipeline durability)
        (optional)
        Refer to :func:`~property_strategies <property_strategies>`.
    :arg dict filter-by-name-wildcard: Enable filter by name with wildcards.
        Requires the :jenkins-plugins:`SCM API Plugin <scm-api>`.

        * **includes** ('str'): Space-separated list
            of name patterns to consider. You may use * as a wildcard;
            for example: `master release*`
        * **excludes** ('str'): Name patterns to
            ignore even if matched by the includes list.
            For example: `release*`

    :extensions:

        * **clean** (`dict`)
            * **after** (`dict`) - Clean the workspace after checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                    submodules and any other subdirectories which contain .git directories
                    (default false)
            * **before** (`dict`) - Clean the workspace before checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                    submodules and any other subdirectories which contain .git directories
                    (default false)
        * **prune** (`bool`) - Prune remote branches (default false)
        * **shallow-clone** (`bool`) - Perform shallow clone (default false)
        * **sparse-checkout** (dict)
            * **paths** (list) - List of paths to sparse checkout. (optional)
        * **depth** (`int`) - Set shallow clone depth (default 1)
        * **do-not-fetch-tags** (`bool`) - Perform a clone without tags
            (default false)
        * **disable-pr-notifications** (`bool`) - Disable default github status
            notifications on pull requests (default false) (Requires the
            :jenkins-plugins:`GitHub Branch Source Plugin
            <disable-github-multibranch-status>`.)
        * **refspecs** (`list(str)`): Which refspecs to fetch.
        * **submodule** (`dict`)
            * **disable** (`bool`) - By disabling support for submodules you
              can still keep using basic git plugin functionality and just have
              Jenkins to ignore submodules completely as if they didn't exist.
            * **recursive** (`bool`) - Retrieve all submodules recursively
              (uses '--recursive' option which requires git>=1.6.5)
            * **tracking** (`bool`) - Retrieve the tip of the configured
              branch in .gitmodules (Uses '\-\-remote' option which requires
              git>=1.8.2)
            * **parent-credentials** (`bool`) - Use credentials from default
              remote of parent repository (default false).
            * **reference-repo** (`str`) - Path of the reference repo to use
              during clone (optional)
            * **timeout** (`int`) - Specify a timeout (in minutes) for
              submodules operations (default 10).
        * **timeout** (`str`) - Timeout for git commands in minutes (optional)
        * **use-author** (`bool`): Use author rather than committer in Jenkin's
            build changeset (default false)
        * **wipe-workspace** (`bool`) - Wipe out workspace before build
            (default true)
        * **lfs-pull** (`bool`) - Call git lfs pull after checkout
            (default false)

    Minimal Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_github_minimal.yaml

    Full Example:

    .. literalinclude::
       /../../tests/multibranch/fixtures/scm_github_full.yaml
    """
    github_path = "org.jenkinsci.plugins.github_branch_source"
    github_path_dscore = "org.jenkinsci.plugins.github__branch__source"

    source = XML.SubElement(
        xml_parent,
        "source",
        {
            "class": "".join([github_path, ".GitHubSCMSource"]),
            "plugin": "github-branch-source",
        },
    )
    mapping = [
        ("", "id", "-".join(["gh", data.get("repo-owner", ""), data.get("repo", "")])),
        ("repo-owner", "repoOwner", None),
        ("repo", "repository", None),
    ]
    helpers.convert_mapping_to_xml(source, data, mapping, fail_required=True)

    mapping_optional = [
        ("api-uri", "apiUri", None),
        ("credentials-id", "credentialsId", None),
    ]
    helpers.convert_mapping_to_xml(source, data, mapping_optional, fail_required=False)

    traits = XML.SubElement(source, "traits")

    # no-pr value is assumed if branch-discovery not mentioned.
    if data.get("branch-discovery", "no-pr"):
        bd = XML.SubElement(
            traits, "".join([github_path_dscore, ".BranchDiscoveryTrait"])
        )
        bd_strategy = {"no-pr": "1", "only-pr": "2", "all": "3"}
        bd_mapping = [("branch-discovery", "strategyId", "no-pr", bd_strategy)]
        helpers.convert_mapping_to_xml(bd, data, bd_mapping, fail_required=True)

    if data.get("ssh-checkout", None):
        cossh = XML.SubElement(
            traits, "".join([github_path_dscore, ".SSHCheckoutTrait"])
        )
        if not isinstance(data.get("ssh-checkout"), bool):
            cossh_credentials = [("credentials", "credentialsId", "")]
            helpers.convert_mapping_to_xml(
                cossh, data.get("ssh-checkout"), cossh_credentials, fail_required=True
            )

    if data.get("discover-tags", False):
        XML.SubElement(traits, "".join([github_path_dscore, ".TagDiscoveryTrait"]))

    if data.get("discover-pr-forks-strategy", "merged-current"):
        dprf = XML.SubElement(
            traits, "".join([github_path_dscore, ".ForkPullRequestDiscoveryTrait"])
        )
        dprf_strategy = {"merge-current": "1", "current": "2", "both": "3"}
        dprf_mapping = [
            ("discover-pr-forks-strategy", "strategyId", "merge-current", dprf_strategy)
        ]
        helpers.convert_mapping_to_xml(dprf, data, dprf_mapping, fail_required=True)

        trust = data.get("discover-pr-forks-trust", "contributors")
        trust_map = {
            "contributors": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustContributors"]
            ),
            "everyone": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustEveryone"]
            ),
            "permission": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustPermission"]
            ),
            "nobody": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustNobody"]
            ),
        }
        if trust not in trust_map:
            raise InvalidAttributeError(
                "discover-pr-forks-trust", trust, trust_map.keys()
            )
        XML.SubElement(dprf, "trust").attrib["class"] = trust_map[trust]

    dpro_strategy = data.get("discover-pr-origin", "merge-current")
    if dpro_strategy:
        dpro = XML.SubElement(
            traits, "".join([github_path_dscore, ".OriginPullRequestDiscoveryTrait"])
        )
        dpro_strategy_map = {"merge-current": "1", "current": "2", "both": "3"}
        if dpro_strategy not in dpro_strategy_map:
            raise InvalidAttributeError(
                "discover-pr-origin", dpro_strategy, dpro_strategy_map.keys()
            )
        dpro_mapping = [
            ("discover-pr-origin", "strategyId", "merge-current", dpro_strategy_map)
        ]
        helpers.convert_mapping_to_xml(dpro, data, dpro_mapping, fail_required=True)

    if data.get("head-filter-regex", None):
        rshf = XML.SubElement(traits, "jenkins.scm.impl.trait.RegexSCMHeadFilterTrait")
        XML.SubElement(rshf, "regex").text = data.get("head-filter-regex")

    if data.get("head-pr-filter-behaviors", None):
        add_filter_branch_pr_behaviors(traits, data.get("head-pr-filter-behaviors"))

    if data.get("property-strategies", None):
        property_strategies(xml_parent, data)

    if data.get("build-strategies", None):
        build_strategies(xml_parent, data)

    add_github_checks_traits(traits, data)
    add_notification_context_trait(traits, data)
    add_filter_by_name_wildcard_behaviors(traits, data)

    # handle the default git extensions like:
    # - clean
    # - shallow-clone
    # - timeout
    # - do-not-fetch-tags
    # - submodule
    # - prune
    # - wipe-workspace
    # - use-author
    # - lfs-pull
    git_extensions(traits, data)

    if data.get("refspecs"):
        refspec_trait = XML.SubElement(
            traits,
            "jenkins.plugins.git.traits.RefSpecsSCMSourceTrait",
            {"plugin": "git"},
        )
        templates = XML.SubElement(refspec_trait, "templates")
        refspecs = data.get("refspecs")
        for refspec in refspecs:
            e = XML.SubElement(
                templates,
                (
                    "jenkins.plugins.git.traits"
                    ".RefSpecsSCMSourceTrait_-RefSpecTemplate"
                ),
            )
            XML.SubElement(e, "value").text = refspec

    # github-only extensions
    disable_github_status_path_dscore = (
        "com.adobe.jenkins.disable__github__multibranch__status"
    )
    if data.get("disable-pr-notifications", False):
        XML.SubElement(
            traits,
            "".join([disable_github_status_path_dscore, ".DisableStatusUpdateTrait"]),
            {"plugin": "disable-github-multibranch-status"},
        )


def build_strategies(xml_parent, data):
    """Configure Basic Branch Build Strategies.

    Requires the :jenkins-plugins:`Basic Branch Build Strategies Plugin
    <basic-branch-build-strategies>`.

    Other build strategies can be configured via raw XML injection.

    :arg list build-strategies: Definition of build strategies.

        * **all-strategies-match** (dict): All sub strategies must match for
            this strategy to match.
            * **strategies** (list): Sub strategies
        * **any-strategies-match** (dict): Builds whenever any of the sub
            strategies match.
            * **strategies** (list): Sub strategies
        * **tags** (dict): Builds tags
            * **ignore-tags-newer-than** (int) The number of days since the tag
                was created before it is eligible for automatic building.
                (optional, default -1)
            * **ignore-tags-older-than** (int) The number of days since the tag
                was created after which it is no longer eligible for automatic
                building. (optional, default -1)
        * **change-request** (dict): Builds change requests / pull requests
            * **ignore-target-only-changes** (bool) Ignore rebuilding merge
                branches when only the target branch changed.
                (optional, default false)
        * **regular-branches** (bool): Builds regular branches whenever a
            change is detected. (optional, default None)
        * **skip-initial-build** (bool): Skip initial build on first branch
            indexing. (optional, default None)
        * **named-branches** (list): Builds named branches whenever a change
          is detected.

            * **exact-name** (dict) Matches the name verbatim.
                * **name** (str) The name to match. (optional)
                * **case-sensitive** (bool) Check this box if the name should
                    be matched case sensitively. (default false)
            * **regex-name** (dict) Matches the name against a regular
              expression.

                * **regex** (str) A Java regular expression to restrict the
                    names. Names that do not match the supplied regular
                    expression will be ignored. (default `^.*$`)
                * **case-sensitive** (bool) Check this box if the name should
                    be matched case sensitively. (default false)
            * **wildcards-name** (dict) Matches the name against an
              include/exclude set of wildcards.

                * **includes** (str) Space-separated list of name patterns to
                    consider. You may use `*` as a wildcard;
                    for example: `master release*` (default `*`)
                * **excludes** (str) Name patterns to ignore even if matched
                    by the includes list. For example: release (optional)
        * **raw** (dict): Injects raw BuildStrategy XML to use other build
            strategy plugins.

    """

    basic_build_strategies = "jenkins.branch.buildstrategies.basic"
    if data.get("build-strategies", None):
        bbs_data = data.get("build-strategies", None)
        bbs = XML.SubElement(xml_parent, "buildStrategies")
    else:
        bbs_data = data.get("strategies", None)
        bbs = XML.SubElement(xml_parent, "strategies")
    for bbs_list in bbs_data:
        if "all-strategies-match" in bbs_list:
            all_elem = XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".AllBranchBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )
            build_strategies(all_elem, bbs_list["all-strategies-match"])

        if "any-strategies-match" in bbs_list:
            any_elem = XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".AnyBranchBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )
            build_strategies(any_elem, bbs_list["any-strategies-match"])

        if "tags" in bbs_list:
            tags = bbs_list["tags"]
            tags_elem = XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".TagBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )

            newer_than = -1
            if "ignore-tags-newer-than" in tags and tags["ignore-tags-newer-than"] >= 0:
                newer_than = str(tags["ignore-tags-newer-than"] * 86400000)
            XML.SubElement(tags_elem, "atLeastMillis").text = str(newer_than)

            older_than = -1
            if "ignore-tags-older-than" in tags and tags["ignore-tags-older-than"] >= 0:
                older_than = str(tags["ignore-tags-older-than"] * 86400000)
            XML.SubElement(tags_elem, "atMostMillis").text = str(older_than)

        if bbs_list.get("regular-branches", False):
            XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".BranchBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )

        if bbs_list.get("skip-initial-build", False):
            XML.SubElement(
                bbs,
                "".join(
                    [basic_build_strategies, ".SkipInitialBuildOnFirstBranchIndexing"]
                ),
                {"plugin": "basic-branch-build-strategies"},
            )

        if "change-request" in bbs_list:
            cr = bbs_list["change-request"]
            cr_elem = XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".ChangeRequestBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )
            itoc = cr.get("ignore-target-only-changes", False)
            XML.SubElement(cr_elem, "ignoreTargetOnlyChanges").text = str(itoc).lower()

        if "named-branches" in bbs_list:
            named_branch_elem = XML.SubElement(
                bbs,
                "".join([basic_build_strategies, ".NamedBranchBuildStrategyImpl"]),
                {"plugin": "basic-branch-build-strategies"},
            )

            filters = XML.SubElement(named_branch_elem, "filters")

            for nb in bbs_list["named-branches"]:
                if "exact-name" in nb:
                    exact_name_elem = XML.SubElement(
                        filters,
                        "".join(
                            [
                                basic_build_strategies,
                                ".NamedBranchBuildStrategyImpl",
                                "_-ExactNameFilter",
                            ]
                        ),
                    )
                    exact_name_mapping = [
                        ("name", "name", ""),
                        ("case-sensitive", "caseSensitive", False),
                    ]
                    helpers.convert_mapping_to_xml(
                        exact_name_elem,
                        nb["exact-name"],
                        exact_name_mapping,
                        fail_required=False,
                    )

                if "regex-name" in nb:
                    regex_name_elem = XML.SubElement(
                        filters,
                        "".join(
                            [
                                basic_build_strategies,
                                ".NamedBranchBuildStrategyImpl",
                                "_-RegexNameFilter",
                            ]
                        ),
                    )
                    regex_name_mapping = [
                        ("regex", "regex", "^.*$"),
                        ("case-sensitive", "caseSensitive", False),
                    ]
                    helpers.convert_mapping_to_xml(
                        regex_name_elem,
                        nb["regex-name"],
                        regex_name_mapping,
                        fail_required=False,
                    )

                if "wildcards-name" in nb:
                    wildcards_name_elem = XML.SubElement(
                        filters,
                        "".join(
                            [
                                basic_build_strategies,
                                ".NamedBranchBuildStrategyImpl",
                                "_-WildcardsNameFilter",
                            ]
                        ),
                    )
                    wildcards_name_mapping = [
                        ("includes", "includes", "*"),
                        ("excludes", "excludes", ""),
                    ]
                    helpers.convert_mapping_to_xml(
                        wildcards_name_elem,
                        nb["wildcards-name"],
                        wildcards_name_mapping,
                        fail_required=False,
                    )

        if "raw" in bbs_list:
            raw_xml = XML.fromstring(bbs_list["raw"].get("xml"))
            remove_ignorable_whitespace(raw_xml)
            XML.SubElement(bbs, None).append(raw_xml)


def property_strategies(xml_parent, data):
    """Configure Basic Branch Property Strategies.

    Requires the :jenkins-plugins:`Branch API Plugin <branch-api>`.

    :arg dict property-strategies: Definition of property strategies.  Either
        `named-branches` or `all-branches` may be specified, but not both.

        * **all-branches** (list): A list of property strategy definitions
            for use with all branches.

            * **suppress-scm-triggering** (dict):
                Suppresses automatic SCM triggering (optional).

                * **branch-regex** (str):
                    Regex matching branch names.
                    Only these branches will be affected by the selected
                    suppression strategy. Default ``^$``.
                * **suppress-strategy** (str):
                    Select what to suppress for branches matching the regex.
                    Valid values: ``suppress-nothing`` (default),
                    ``suppress-branch-indexing``, or ``suppress-webhooks``.

                    .. note::
                        Using ``suppress-scm-triggering: true`` is deprecated
                        since Branch API 2.1045.v4ec3ed07b_e4f.
            * **pipeline-branch-durability-override** (str): Set a custom
                branch speed/durability level. Valid values:
                performance-optimized, survivable-nonatomic, or
                max-survivability (optional)
                Requires the :jenkins-plugins:`Pipeline Multibranch Plugin
                <workflow-multibranch>`
            * **trigger-build-on-pr-comment** (str or dict): The comment body to
                trigger a new build for a PR job when it is received. This
                is compiled as a case-insensitive regular expression, so
                use ``".*"`` to trigger a build on any comment whatsoever.
                (optional)
                If dictionary syntax is used, the option requires 2 fields:
                ``comment`` with the comment body and ``allow-untrusted-users``
                (bool) causing the plugin to skip checking if the comment author
                is a collaborator of the GitHub project.
                Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                <github-pr-comment-build>`
            * **trigger-build-on-pr-review** (bool or dict): This property will
                cause a job for a pull request ``(PR-*)`` to be triggered
                immediately when a review is made on the PR in GitHub.
                This has no effect on jobs that are not for pull requests.
                (optional)
                If dictionary syntax is used, the option requires
                ``allow-untrusted-users`` (bool) causing the plugin to skip
                checking if the review author is a collaborator of the GitHub
                project.
                Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                <github-pr-comment-build>`
            * **trigger-build-on-pr-update** (bool or dict): This property will
                cause a job for a pull request ``(PR-*)`` to be triggered
                immediately when the PR title or description is edited in
                GitHub. This has no effect on jobs that are not for pull
                requests. (optional)
                If dictionary syntax is used, the option requires
                ``allow-untrusted-users`` (bool) causing the plugin to skip
                checking if the update author is a collaborator of the GitHub
                project.
                Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                <github-pr-comment-build>`
        * **named-branches** (dict): Named branches get different properties.
            Comprised of a list of defaults and a list of property strategy
            exceptions for use with specific branches.

            * **defaults** (list): A list of property strategy definitions
                to be applied by default to all branches, unless overridden
                by an entry in `exceptions`

                * **suppress-scm-triggering** (dict):
                    Suppresses automatic SCM triggering (optional).

                    * **branch-regex** (str):
                        Regex matching branch names.
                        Only these branches will be affected by the selected
                        suppression strategy. Default ``^$``.
                    * **suppress-strategy** (str):
                        Select what to suppress for branches matching the regex.
                        Valid values: ``suppress-nothing`` (default),
                        ``suppress-branch-indexing``, or ``suppress-webhooks``.

                        .. note::
                            Using ``suppress-scm-triggering: true`` is deprecated
                            since Branch API 2.1045.v4ec3ed07b_e4f.
                * **pipeline-branch-durability-override** (str): Set a custom
                    branch speed/durability level. Valid values:
                    performance-optimized, survivable-nonatomic, or
                    max-survivability (optional)
                    Requires the :jenkins-plugins:`Pipeline Multibranch Plugin
                    <workflow-multibranch>`
                * **trigger-build-on-pr-comment** (str or dict): The comment body to
                    trigger a new build for a PR job when it is received. This
                    is compiled as a case-insensitive regular expression, so
                    use ``".*"`` to trigger a build on any comment whatsoever.
                    (optional)
                    If dictionary syntax is used, the option accepts 2 fields:
                    ``comment`` (str, required) with the comment body and
                    ``allow-untrusted-users`` (bool, optional) causing the plugin
                    to skip checking if the comment author is a collaborator of
                    the GitHub project.
                    Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                    <github-pr-comment-build>`
                * **trigger-build-on-pr-review** (bool or dict): This property will
                    cause a job for a pull request ``(PR-*)`` to be triggered
                    immediately when a review is made on the PR in GitHub.
                    This has no effect on jobs that are not for pull requests.
                    (optional)
                    If dictionary syntax is used, the option requires
                    ``allow-untrusted-users`` (bool) causing the plugin to skip
                    checking if the review author is a collaborator of the GitHub
                    project.
                    Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                    <github-pr-comment-build>`
                * **trigger-build-on-pr-update** (bool or dict): This property will
                    cause a job for a pull request ``(PR-*)`` to be triggered
                    immediately when the PR title or description is edited in
                    GitHub. This has no effect on jobs that are not for pull
                    requests. (optional)
                    If dictionary syntax is used, the option requires
                    ``allow-untrusted-users`` (bool) causing the plugin to skip
                    checking if the update author is a collaborator of the GitHub
                    project.
                    Requires the :jenkins-plugins:`GitHub PR Comment Build Plugin
                    <github-pr-comment-build>`

            * **exceptions** (list): A list of branch names and the property
                strategies to be used on that branch, instead of any listed
                in `defaults`.

                * **exception** (dict): Defines exception
                    * **branch-name** (str): Name of the branch to which these
                        properties will be applied.
                    * **properties** (list): A list of properties to apply to
                        this branch.

                        * **suppress-scm-triggering** (dict):
                            Suppresses automatic SCM triggering (optional).

                            * **branch-regex** (str):
                                Regex matching branch names.
                                Only these branches will be affected by the selected
                                suppression strategy. Default ``^$``.
                            * **suppress-strategy** (str):
                                Select what to suppress for branches matching the regex.
                                Valid values: ``suppress-nothing`` (default),
                                ``suppress-branch-indexing``, or ``suppress-webhooks``.

                                .. note::
                                    Using ``suppress-scm-triggering: true`` is deprecated
                                    since Branch API 2.1045.v4ec3ed07b_e4f.

                        * **pipeline-branch-durability-override** (str): Set a
                            custom branch speed/durability level. Valid values:
                            performance-optimized, survivable-nonatomic, or
                            max-survivability (optional)
                            Requires the :jenkins-plugins:`Pipeline
                            Multibranch Plugin <workflow-multibranch>`
    """

    valid_prop_strats = ["all-branches", "named-branches"]

    basic_property_strategies = "jenkins.branch"

    prop_strats = data.get("property-strategies", None)

    if prop_strats:

        for prop_strat in prop_strats:
            if prop_strat not in valid_prop_strats:
                raise InvalidAttributeError(
                    "property-strategies", prop_strat, valid_prop_strats
                )
        if len(prop_strats) > 1:
            raise JenkinsJobsException("Only one property strategy may be specified")

        all_branches = prop_strats.get("all-branches", None)
        named_branches = prop_strats.get("named-branches", None)

        if all_branches:

            strat_elem = XML.SubElement(
                xml_parent,
                "strategy",
                {
                    "class": "".join(
                        [basic_property_strategies, ".DefaultBranchPropertyStrategy"]
                    )
                },
            )
            props_elem = XML.SubElement(
                strat_elem, "properties", {"class": "java.util.Arrays$ArrayList"}
            )
            props_elem = XML.SubElement(
                props_elem,
                "a",
                {
                    "class": "".join(
                        [basic_property_strategies, ".BranchProperty-array"]
                    )
                },
            )

            apply_property_strategies(props_elem, all_branches)

        elif named_branches:

            strat_elem = XML.SubElement(
                xml_parent,
                "strategy",
                {
                    "class": "".join(
                        [
                            basic_property_strategies,
                            ".NamedExceptionsBranchPropertyStrategy",
                        ]
                    )
                },
            )

            nbs_defaults = named_branches.get("defaults", None)
            if nbs_defaults:

                props_elem = XML.SubElement(
                    strat_elem,
                    "defaultProperties",
                    {"class": "java.util.Arrays$ArrayList"},
                )
                props_elem = XML.SubElement(
                    props_elem,
                    "a",
                    {
                        "class": "".join(
                            [basic_property_strategies, ".BranchProperty-array"]
                        )
                    },
                )

                apply_property_strategies(props_elem, nbs_defaults)

            nbs_exceptions = named_branches.get("exceptions", None)
            if nbs_exceptions:

                props_elem = XML.SubElement(
                    strat_elem,
                    "namedExceptions",
                    {"class": "java.util.Arrays$ArrayList"},
                )
                props_elem = XML.SubElement(
                    props_elem,
                    "a",
                    {
                        "class": "".join(
                            [
                                basic_property_strategies,
                                ".NamedExceptionsBranchPropertyStrategy$Named-array",
                            ]
                        )
                    },
                )

                for named_exception in nbs_exceptions:
                    named_exception = named_exception.get("exception", None)
                    if not named_exception:
                        continue

                    exc_elem = XML.SubElement(
                        props_elem,
                        "".join(
                            [
                                basic_property_strategies,
                                ".NamedExceptionsBranchPropertyStrategy_-Named",
                            ]
                        ),
                    )

                    ne_branch_name = named_exception.get("branch-name", None)
                    if ne_branch_name is not None:
                        XML.SubElement(exc_elem, "name").text = ne_branch_name

                    ne_properties = named_exception.get("properties", None)
                    if ne_properties:
                        exc_elem = XML.SubElement(
                            exc_elem, "props", {"class": "java.util.Arrays$ArrayList"}
                        )
                        exc_elem = XML.SubElement(
                            exc_elem,
                            "a",
                            {
                                "class": "".join(
                                    [basic_property_strategies, ".BranchProperty-array"]
                                )
                            },
                        )
                        apply_property_strategies(exc_elem, ne_properties)


def apply_property_strategies(props_elem, props_list):
    # there are 3 locations at which property strategies can be defined:
    # globally (all-branches), defaults (named-branches), exceptions
    # (also named-branches)

    basic_property_strategies = "jenkins.branch"
    workflow_multibranch = "org.jenkinsci.plugins.workflow.multibranch"
    pr_comment_build = "com.adobe.jenkins.github__pr__comment__build"
    # Valid options for the pipeline branch durability override.
    pbdo_map = collections.OrderedDict(
        [
            ("max-survivability", "MAX_SURVIVABILITY"),
            ("performance-optimized", "PERFORMANCE_OPTIMIZED"),
            ("survivable-nonatomic", "SURVIVABLE_NONATOMIC"),
        ]
    )

    pcb_bool_opts = collections.OrderedDict(
        [
            ("trigger-build-on-pr-review", ".TriggerPRReviewBranchProperty"),
            ("trigger-build-on-pr-update", ".TriggerPRUpdateBranchProperty"),
        ]
    )

    # valid options for suppress SCM trigger property
    # strategy name says what's blocked
    sst_map = collections.OrderedDict(
        [
            ("suppress-nothing", "NONE"),
            ("suppress-webhooks", "EVENTS"),
            ("suppress-branch-indexing", "INDEXING"),
        ]
    )

    for dbs_list in props_list:

        sst_val = dbs_list.get("suppress-scm-triggering", False)
        if sst_val:
            sst_elem = XML.SubElement(
                props_elem,
                "".join([basic_property_strategies, ".NoTriggerBranchProperty"]),
            )

            if isinstance(sst_val, dict):
                if "branch-regex" not in sst_val:
                    raise MissingAttributeError("suppress-scm-triggering[branch-regex]")
                if "suppression-strategy" not in sst_val:
                    raise MissingAttributeError(
                        "suppress-scm-triggering[suppression-strategy]"
                    )

                sst_ss_val = sst_val["suppression-strategy"]
                if not sst_map.get(sst_ss_val):
                    raise InvalidAttributeError(
                        "suppression-strategy", sst_ss_val, sst_map.keys()
                    )
                XML.SubElement(sst_elem, "triggeredBranchesRegex").text = sst_val[
                    "branch-regex"
                ]
                XML.SubElement(sst_elem, "strategy").text = sst_map[sst_ss_val]

            elif isinstance(sst_val, bool):
                # no sub-elements in this case
                pass
            else:
                raise InvalidAttributeError("suppress-scm-triggering", sst_val)

        pbdo_val = dbs_list.get("pipeline-branch-durability-override", None)
        if pbdo_val:
            if not pbdo_map.get(pbdo_val):
                raise InvalidAttributeError(
                    "pipeline-branch-durability-override", pbdo_val, pbdo_map.keys()
                )
            pbdo_elem = XML.SubElement(
                props_elem,
                "".join([workflow_multibranch, ".DurabilityHintBranchProperty"]),
                {"plugin": "workflow-multibranch"},
            )
            XML.SubElement(pbdo_elem, "hint").text = pbdo_map.get(pbdo_val)

        tbopc_val = dbs_list.get("trigger-build-on-pr-comment", None)
        if tbopc_val:
            tbopc_elem = XML.SubElement(
                props_elem,
                "".join([pr_comment_build, ".TriggerPRCommentBranchProperty"]),
                {"plugin": "github-pr-comment-build"},
            )
            if isinstance(tbopc_val, dict):
                if "comment" not in tbopc_val:
                    raise MissingAttributeError(
                        "trigger-build-on-pr-comment[comment]",
                        pos=dbs_list.key_pos.get("trigger-build-on-pr-comment"),
                    )
                XML.SubElement(tbopc_elem, "commentBody").text = tbopc_val["comment"]
                if tbopc_val.get("allow-untrusted-users", False):
                    XML.SubElement(tbopc_elem, "allowUntrusted").text = "true"
            elif isinstance(tbopc_val, str):
                XML.SubElement(tbopc_elem, "commentBody").text = tbopc_val
            else:
                attr = "trigger-build-on-pr-comment"
                ctx = [Context(f"For attribute {attr!r}", dbs_list.key_pos.get(attr))]
                raise InvalidAttributeError(
                    attr,
                    tbopc_val,
                    pos=dbs_list.value_pos.get(attr),
                    ctx=ctx,
                )
        for opt in pcb_bool_opts:
            opt_value = dbs_list.get(opt, None)
            if opt_value:
                opt_elem = XML.SubElement(
                    props_elem,
                    "".join([pr_comment_build, pcb_bool_opts.get(opt)]),
                    {"plugin": "github-pr-comment-build"},
                )
                if isinstance(opt_value, dict):
                    if opt_value.get("allow-untrusted-users", False):
                        XML.SubElement(opt_elem, "allowUntrusted").text = "true"
                elif isinstance(opt_value, bool):
                    # no sub-elements in this case
                    pass
                else:
                    ctx = Context(f"For attribute {opt!r}", dbs_list.key_pos.get(opt))
                    raise InvalidAttributeError(
                        opt, opt_value, pos=dbs_list.value_pos.get(opt), ctx=[ctx]
                    )


def add_filter_branch_pr_behaviors(traits, data):
    """Configure Filter Branch PR behaviors

    Requires the :jenkins-plugins:`SCM Filter Branch PR Plugin
    <scm-filter-branch-pr>`.

    :arg list head-pr-filter-behaviors: Definition of filters.

        * **head-pr-destined-regex** (dict): Filter by name incl. PR destined to
            this branch with regexp

            * **branch-regexp** (str) Regular expression to filter branches and
                PRs (optional, default ".*")
            * **tag-regexp** (str) Regular expression to filter tags
                (optional, default "(?!.*)")

        * **head-pr-destined-wildcard** (dict): Filter by name incl. PR
            destined to this branch with wildcard

            * **branch-includes** (str) Wildcard expression to include branches
                and PRs (optional, default "*")
            * **tag-includes** (str) Wildcard expression to include tags
                (optional, default "")
            * **branch-excludes** (str) Wildcard expression to exclude branches
                and PRs (optional, default "")
            * **tag-excludes** (str) Wildcard expression to exclude tags
                (optional, default "*")

        * **head-pr-originated-regex** (dict): Filter by name incl. PR destined
            to this branch with regexp

            * **branch-regexp** (str) Regular expression to filter branches
                and PRs (optional, default ".*")
            * **tag-regexp** (str) Regular expression to filter tags
                (optional, default "(?!.*)")

        * **head-pr-originated-wildcard** (dict): Filter by name incl. PR
            destined to this branch with wildcard

            * **branch-includes** (str) Wildcard expression to include branches
                and PRs (optional, default "*")
            * **tag-includes** (str) Wildcard expression to include tags
                (optional, default "")
            * **branch-excludes** (str) Wildcard expression to exclude branches
                and PRs (optional, default "")
            * **tag-excludes** (str) Wildcard expression to exclude tags
                (optional, default "*")
    """

    regexp_mapping = [
        ("branch-regexp", "regex", ".*"),
        ("tag-regexp", "tagRegex", "(?!.*)"),
    ]
    wildcard_mapping = [
        ("branch-includes", "includes", "*"),
        ("branch-excludes", "excludes", ""),
        ("tag-includes", "tagIncludes", ""),
        ("tag-excludes", "tagExcludes", "*"),
    ]

    if data.get("head-pr-destined-regex"):
        rshf = XML.SubElement(
            traits,
            "net.gleske.scmfilter.impl.trait.RegexSCMHeadFilterTrait",
            {"plugin": "scm-filter-branch-pr"},
        )
        helpers.convert_mapping_to_xml(
            rshf, data.get("head-pr-destined-regex"), regexp_mapping, fail_required=True
        )

    if data.get("head-pr-destined-wildcard"):
        wshf = XML.SubElement(
            traits,
            "net.gleske.scmfilter.impl.trait.WildcardSCMHeadFilterTrait",
            {"plugin": "scm-filter-branch-pr"},
        )
        helpers.convert_mapping_to_xml(
            wshf,
            data.get("head-pr-destined-wildcard"),
            wildcard_mapping,
            fail_required=True,
        )

    if data.get("head-pr-originated-regex"):
        rsof = XML.SubElement(
            traits,
            "net.gleske.scmfilter.impl.trait.RegexSCMOriginFilterTrait",
            {"plugin": "scm-filter-branch-pr"},
        )
        helpers.convert_mapping_to_xml(
            rsof,
            data.get("head-pr-originated-regex"),
            regexp_mapping,
            fail_required=True,
        )

    if data.get("head-pr-originated-wildcard"):
        wsof = XML.SubElement(
            traits,
            "net.gleske.scmfilter.impl.trait.WildcardSCMOriginFilterTrait",
            {"plugin": "scm-filter-branch-pr"},
        )
        helpers.convert_mapping_to_xml(
            wsof,
            data.get("head-pr-originated-wildcard"),
            wildcard_mapping,
            fail_required=True,
        )


def add_filter_by_name_wildcard_behaviors(traits, data):
    """Configure branch filtering behaviors.

    :arg dict filter-by-name-wildcard: Enable filter by name with wildcards.
        Requires the :jenkins-plugins:`SCM API Plugin <scm-api>`.

        * **includes** ('str'): Space-separated list
            of name patterns to consider. You may use * as a wildcard;
            for example: `master release*`
        * **excludes** ('str'): Name patterns to
            ignore even if matched by the includes list.
            For example: `release*`
    """
    if data.get("filter-by-name-wildcard", None):
        wscmf_name = XML.SubElement(
            traits,
            "jenkins.scm.impl.trait.WildcardSCMHeadFilterTrait",
            {"plugin": "scm-api"},
        )
        wscmf_name_mapping = [
            ("includes", "includes", ""),
            ("excludes", "excludes", ""),
        ]
        helpers.convert_mapping_to_xml(
            wscmf_name,
            data.get("filter-by-name-wildcard", ""),
            wscmf_name_mapping,
            fail_required=True,
        )


def add_notification_context_trait(traits, data):
    """Change the default GitHub check notification context from
    "continuous-integration/jenkins/SUFFIX" to a custom label / suffix.
    (set a label and suffix to true or false, optional)

    Requires the :jenkins-plugins:`Github Custom Notification Context SCM
    Behaviour <github-scm-trait-notification-context>`.

    :arg dict notification-context: Definition of notification-context. A
        `label` must be specified. `suffix` may be specified with true /
        false, default being true.

        * **label** (str): The text of the context label for Github status
            notifications.
        * **suffix** (bool): Appends the relevant suffix to the context label
            based on the build type. '/pr-merge', '/pr-head' or '/branch'
            (optional, default true)
    """
    if data.get("notification-context", None):
        nc_trait = XML.SubElement(
            traits,
            "org.jenkinsci.plugins.githubScmTraitNotificationContext."
            "NotificationContextTrait",
        )
        nc = data.get("notification-context")
        nc_trait_label = XML.SubElement(nc_trait, "contextLabel")
        nc_trait_suffix = XML.SubElement(nc_trait, "typeSuffix")
        if isinstance(nc, str):
            nc_trait_label.text = nc
            nc_trait_suffix.text = "true"
        else:
            nc_trait_label.text = nc.get("label")
            nc_suffix = nc.get("suffix", None)
            if nc_suffix is None:
                nc_trait_suffix.text = "true"
            elif type(nc_suffix) is bool:
                nc_trait_suffix.text = str(nc_suffix).lower()
            else:
                nc_trait_suffix.text = nc_suffix


def add_github_checks_traits(traits, data):
    """Enable and configure the usage of GitHub Checks API

    Requires the :jenkins-plugins:`Github Checks <github-checks>` plugin.

    :arg dict status-checks:

        * **name** (str): The text of the context label for GitHub Checks entry
        * **skip** (bool): Skips publishing Checks (optional, default false)
        * **skip-branch-source-notifications** (bool): Disables the default
            option of publishing statuses through Status API
            (optional, default false)
        * **publish-unstable-as-neutral** (bool): Publishes UNSTABLE builds
            as neutral (not failed) checks
            (optional, default false)
        * **suppress-log-output** (bool): Suppresses sending build logs to GitHub
            (optional, default false)
        * **suppress-progress-updates** (bool): Suppresses updating build progress
            (optional, default false)
        * **verbose-logs** (bool): Enables sending build console logs to GitHub
            (optional, default false)
    """
    if data.get("status-checks", None):
        status_checks_section = data["status-checks"]
        if status_checks_section.get("verbose-logs", None):
            sct = XML.SubElement(
                traits,
                "io.jenkins.plugins.checks.github.config.GitHubSCMSourceChecksTrait",
                {"plugin": "github-checks"},
            )
            sct_mapping = [("verbose-logs", "verboseConsoleLog", "false")]
            helpers.convert_mapping_to_xml(
                sct, status_checks_section, sct_mapping, fail_required=True
            )

        status_check_trait = XML.SubElement(
            traits,
            "io.jenkins.plugins.checks.github.status.GitHubSCMSourceStatusChecksTrait",
            {"plugin": "github-checks"},
        )
        status_check_trait_mapping = [
            ("skip", "skip", "false"),
            ("skip-branch-source-notifications", "skipNotifications", "false"),
            ("publish-unstable-as-neutral", "unstableBuildNeutral", "false"),
            ("name", "name", "Jenkins"),
            ("suppress-log-output", "suppressLogs", "false"),
            ("suppress-progress-updates", "skipProgressUpdates", "false"),
        ]
        helpers.convert_mapping_to_xml(
            status_check_trait,
            status_checks_section,
            status_check_trait_mapping,
            fail_required=True,
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import yaml

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from .cached_property import cached_property


LINE_SEPARATORS = "\0\r\n\x85\u2028\u2029"
WHITESPACE_CHARS = " \t"


class Pos:
    @classmethod
    def from_node(cls, node, line_ofs=0, column_ofs=0):
        mark = node.start_mark
        return cls(mark, mark.name, mark.line + line_ofs, mark.column + column_ofs)

    @classmethod
    def from_file(cls, path, text):
        mark = yaml.Mark(str(path), 0, 0, 0, text, 0)
        return cls(mark, path, 0, 0)

    def __init__(self, mark, path, line, column):
        self._mark = mark
        self.path = path
        self.line = line  # Starts from 0.
        self.column = column  # Starts from 0.

    def __repr__(self):
        return f"<Pos {self.path}:{self.line}:{self.column}>"

    def with_offset(self, line_ofs=0, column_ofs=0):
        line_ptr = self._move_ptr_by_lines(line_ofs)
        ptr = line_ptr + column_ofs
        mark = self._clone_mark(self._mark, ptr)
        if line_ofs:
            column = column_ofs  # Start from new line.
        else:
            column = self.column + column_ofs
        return Pos(mark, self.path, self.line + line_ofs, column)

    def with_contents_start(self):
        ptr = self._mark.pointer
        buf = self._mark.buffer
        while (
            ptr < len(buf)
            and buf[ptr] not in LINE_SEPARATORS
            and buf[ptr] in WHITESPACE_CHARS
        ):
            ptr += 1
        mark = self._clone_mark(self._mark, ptr)
        return Pos(mark, self.path, self.line, self.column + ptr - self._mark.pointer)

    @cached_property
    def snippet(self):
        return self._mark.get_snippet(max_length=100)

    @cached_property
    def body(self):
        return self._mark.buffer[self._mark.pointer :]

    def _clone_mark(self, mark, ptr):
        return yaml.Mark(
            mark.name,
            mark.index,
            mark.line,
            mark.column,
            mark.buffer,
            ptr,
        )

    def _move_ptr_by_lines(self, line_ofs):
        ptr = self._mark.pointer
        buf = self._mark.buffer
        while line_ofs > 0 and ptr < len(buf):
            if buf[ptr] in LINE_SEPARATORS:
                line_ofs -= 1
            ptr += 1
        return ptr



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
from collections import namedtuple

from .errors import Context, JenkinsJobsException
from .loc_loader import LocList, LocDict
from jenkins_jobs.expander import YamlObjectsExpander


Dimension = namedtuple("Dimension", "axis params")


def _decode_axis_value(axis, value, key_pos, value_pos):
    if not isinstance(value, (list, LocList)):
        yield {axis: value}
        return
    for idx, item in enumerate(value):
        if not isinstance(item, (dict, LocDict)):
            d = LocDict()
            if type(value) is LocList:
                d.set_item(axis, item, key_pos, value.value_pos[idx])
            else:
                d[axis] = item
            yield d
            continue
        if len(item.items()) != 1:
            raise JenkinsJobsException(
                f"Expected a value or a dict with single element, but got: {item!r}",
                pos=value.value_pos[idx],
                ctx=[Context(f"In parameter {axis!r} definition", key_pos)],
            )
        value, p = next(iter(item.items()))
        yield LocDict.merge(
            {axis: value},  # Point axis value.
            p,  # Point-specific parameters. May override axis value.
        )


def enum_dimensions_params(axes, params, defaults):
    expander = YamlObjectsExpander()
    if not axes:
        # No axes - instantiate one job/view.
        yield {}
        return
    dim_values = []
    for axis in axes:
        try:
            value, key_pos, value_pos = params.item_with_pos(axis)
        except KeyError:
            try:
                value = defaults[axis]
                key_pos = value_pos = None
            except KeyError:
                continue  # May be, value would be received from an another axis values.
        expanded_value = expander.expand(value, params, key_pos, value_pos)
        value = [
            Dimension(axis, params)
            for params in _decode_axis_value(axis, expanded_value, key_pos, value_pos)
        ]
        dim_values.append(value)
    for dimensions in itertools.product(*dim_values):
        overrides = {}  # Axis -> overridden param.
        for dim in dimensions:
            for name, value in dim.params.items():
                if name != dim.axis:
                    overrides[name] = value
        param_dicts = [d.params for d in dimensions]
        # Parameter overrides should take precedence over axis values.
        yield LocDict.merge(*param_dicts, overrides)


def _match_exclude(params, exclude, pos):
    if not isinstance(exclude, dict):
        raise JenkinsJobsException(
            f"Expected a dict, but got: {exclude!r}",
            pos=pos,
        )
    if not exclude:
        raise JenkinsJobsException(
            f"Expected a dict, but is empty: {exclude!r}",
            pos=pos,
        )
    for axis, value in exclude.items():
        try:
            v = params[axis]
        except KeyError:
            raise JenkinsJobsException(
                f"Unknown axis {axis!r}",
                pos=pos,
            )
        if value != v:
            return False
    # All required exclude values are matched.
    return True


def is_point_included(exclude_list, params, key_pos=None):
    if not exclude_list:
        return True
    try:
        for idx, exclude in enumerate(exclude_list):
            if _match_exclude(params, exclude, exclude_list.value_pos[idx]):
                return False
    except JenkinsJobsException as x:
        raise x.with_context(
            "In template exclude list",
            pos=key_pos,
        )
    return True



#!/usr/bin/env python
# Copyright (C) 2015 Wayne Warren
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import fnmatch
import logging
import time

from jenkins_jobs.builder import JenkinsManager
from jenkins_jobs.registry import ModuleRegistry
from jenkins_jobs.roots import Roots
from jenkins_jobs.xml_config import XmlJobGenerator
from jenkins_jobs.xml_config import XmlViewGenerator
from jenkins_jobs.loader import load_files


logger = logging.getLogger(__name__)


def matches(name, glob_list):
    """
    Checks if the given string, ``name``, matches any of the glob patterns in
    the iterable, ``glob_list``

    :arg str name: String (job or view name) to test if it matches a pattern
    :arg iterable glob_list: glob patterns to match (list, tuple, set, etc.)
    """
    return any(fnmatch.fnmatch(name, glob) for glob in glob_list)


def filter_matching(item_list, glob_list):
    if not glob_list:
        return item_list
    return [item for item in item_list if matches(item.name, glob_list)]


class BaseSubCommand(metaclass=abc.ABCMeta):
    """Base class for Jenkins Job Builder subcommands, intended to allow
    subcommands to be loaded as stevedore extensions by third party users.
    """

    def __init__(self):
        pass

    @abc.abstractmethod
    def parse_args(self, subparsers, recursive_parser):
        """Define subcommand arguments.

        :param subparsers
          A sub parser object. Implementations of this method should
          create a new subcommand parser by calling
            parser = subparsers.add_parser('command-name', ...)
          This will return a new ArgumentParser object; all other arguments to
          this method will be passed to the argparse.ArgumentParser constructor
          for the returned object.
        """

    @abc.abstractmethod
    def execute(self, config):
        """Execute subcommand behavior.

        :param config
          JJBConfig object containing final configuration from config files,
          command line arguments, and environment variables.
        """

    @staticmethod
    def parse_option_recursive_exclude(parser):
        """Add '--recursive'  and '--exclude' arguments to given parser."""
        parser.add_argument(
            "-r",
            "--recursive",
            action="store_true",
            dest="recursive",
            default=False,
            help="look for yaml files recursively",
        )

        parser.add_argument(
            "-x",
            "--exclude",
            dest="exclude",
            action="append",
            default=[],
            help="paths to exclude when using recursive search, "
            "uses standard globbing.",
        )


class JobsSubCommand(BaseSubCommand):
    """Base class for Jenkins Job Builder subcommands which generates jobs."""

    def load_roots(self, jjb_config, path_list):
        roots = Roots(jjb_config)
        load_files(jjb_config, roots, path_list)
        return roots

    def make_jobs_and_views_xml(self, jjb_config, path_list, glob_list):
        logger.info("Updating jobs in {0} ({1})".format(path_list, glob_list))
        orig = time.time()

        roots = self.load_roots(jjb_config, path_list)

        builder = JenkinsManager(jjb_config)

        registry = ModuleRegistry(jjb_config, builder.plugins_list)
        registry.set_macros(roots.macros)

        jobs = filter_matching(roots.generate_jobs(), glob_list)
        views = filter_matching(roots.generate_views(), glob_list)

        registry.amend_job_dicts(jobs)

        xml_job_generator = XmlJobGenerator(registry)
        xml_view_generator = XmlViewGenerator(registry)

        xml_jobs = xml_job_generator.generateXML(jobs)
        xml_views = xml_view_generator.generateXML(views)

        step = time.time()
        logging.debug("%d XML files generated in %ss", len(jobs), str(step - orig))

        return builder, xml_jobs, xml_views



from functools import lru_cache

# cached_property was introduced in Python 3.8.
# TODO: Remove this file when support for Python 3.7 is dropped.
# Recipe from https://stackoverflow.com/a/19979379


def cached_property(fn):
    return property(lru_cache()(fn))



#!/usr/bin/env python
# Copyright (C) 2017 Thanh Ha
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import yaml

from jenkins_jobs.builder import JenkinsManager
import jenkins_jobs.cli.subcommand.base as base


logger = logging.getLogger(__name__)


class GetPluginsInfoSubCommand(base.BaseSubCommand):
    def parse_args(self, subparser):
        plugins_info = subparser.add_parser(
            "get-plugins-info", help="get plugins info yaml by querying Jenkins server."
        )

        plugins_info.add_argument(
            "-o",
            "--output-file",
            default="plugins_info.yaml",
            dest="plugins_info_file",
            help="file to save output to.",
        )

    def execute(self, options, jjb_config):
        builder = JenkinsManager(jjb_config)
        plugin_data = builder.jenkins.get_plugins_info()
        plugins_info = []
        for plugin in plugin_data:
            info = {
                "longName": str(plugin["longName"]),
                "shortName": str(plugin["shortName"]),
                "version": str(plugin["version"]),
            }
            plugins_info.append(info)

        if options.plugins_info_file:
            with open(options.plugins_info_file, "w") as outfile:
                outfile.write(yaml.dump(plugins_info))
            logger.info("Generated {} file".format(options.plugins_info_file))
        else:
            print(yaml.dump(plugins_info))



#!/usr/bin/env python
# Copyright (C) 2012 OpenStack, LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Manage jobs in Jenkins server

import errno
import hashlib
import io
import logging
import os
from pprint import pformat
import re
from six.moves.urllib.parse import quote
import time
import xml.etree.ElementTree as XML

import jenkins

from jenkins_jobs.alphanum import AlphanumSort
from jenkins_jobs.cache import JobCache
from jenkins_jobs.constants import MAGIC_MANAGE_STRING
from jenkins_jobs.parallel import concurrent
from jenkins_jobs import utils

__all__ = ["JenkinsManager"]

logger = logging.getLogger(__name__)

_DEFAULT_TIMEOUT = object()


class JenkinsManager(object):
    def __init__(self, jjb_config):
        url = jjb_config.jenkins["url"]
        user = jjb_config.jenkins["user"]
        password = jjb_config.jenkins["password"]
        timeout = jjb_config.jenkins["timeout"]

        if timeout != _DEFAULT_TIMEOUT:
            self.jenkins = jenkins.Jenkins(url, user, password, timeout)
        else:
            self.jenkins = jenkins.Jenkins(url, user, password)

        self.cache = JobCache(
            jjb_config.jenkins["url"], flush=jjb_config.builder["flush_cache"]
        )

        self._plugins_list = jjb_config.builder["plugins_info"]
        self._jobs = None
        self._job_list = None
        self._views = None
        self._view_list = None
        self._jjb_config = jjb_config

    def _setup_output(self, output, item, config_xml=False):
        output_dir = output
        output_fn = os.path.join(output, item)
        if "/" in item:
            # in item folder
            output_fn = os.path.join(output, os.path.normpath(item))
            output_dir = os.path.dirname(output_fn)

        # if in a folder, re-adding name to the directory here
        if config_xml:
            output_dir = os.path.join(output_dir, os.path.basename(item))
            output_fn = os.path.join(output_dir, "config.xml")

        if output_dir != output:
            logger.debug("Creating directory %s" % output_dir)
            try:
                os.makedirs(output_dir)
            except OSError:
                if not os.path.isdir(output_dir):
                    raise

        return output_fn

    @property
    def jobs(self):
        if self._jobs is None:
            # populate jobs
            self._jobs = self.jenkins.get_all_jobs()

        return self._jobs

    @property
    def job_list(self):
        if self._job_list is None:
            # python-jenkins uses 'fullname' for folder/name combination
            self._job_list = set(job["fullname"] for job in self.jobs)
        return self._job_list

    def _job_format(self, job_name):
        # returns job name or url based on config option
        if self._jjb_config.builder["print_job_urls"]:
            return (
                self._jjb_config.jenkins["url"]
                + "/job/"
                + quote("/job/".join(job_name.split("/")).encode("utf8"))
                + "/"
            )
        else:
            return job_name

    def _view_format(self, view_name):
        # returns job name or url based on config option
        if self._jjb_config.builder["print_job_urls"]:
            parts = view_name.split("/")
            return (
                self._jjb_config.jenkins["url"]
                + "".join(["/job/" + item for item in parts[:-1]])
                + "/view/"
                + parts[-1]
                + "/"
            )
        else:
            return view_name

    def update_job(self, job_name, xml):
        if self.is_job(job_name):
            logger.info(
                "Reconfiguring jenkins job {0}".format(self._job_format(job_name))
            )
            self.jenkins.reconfig_job(job_name, xml)
        else:
            logger.info("Creating jenkins job {0}".format(self._job_format(job_name)))
            self.jenkins.create_job(job_name, xml)

    def is_job(self, job_name, use_cache=True):
        if use_cache:
            if job_name in self.job_list:
                return True

        return self.jenkins.job_exists(job_name)

    def get_job_md5(self, job_name):
        xml = self.jenkins.get_job_config(job_name)
        return hashlib.md5(xml.encode("utf-8")).hexdigest()

    def delete_job(self, job_name):
        if self.is_job(job_name):
            logger.info("Deleting jenkins job {0}".format(job_name))
            self.jenkins.delete_job(job_name)

    def get_plugins_info(self):
        """Return a list of plugin_info dicts.

        Returns one for each plugin on the Jenkins instance.
        """
        try:
            plugins_list = self.jenkins.get_plugins().values()

        except jenkins.JenkinsException as e:
            if re.search("(Connection refused|Forbidden)", str(e)):
                logger.warning(
                    "Unable to retrieve Jenkins Plugin Info from {0},"
                    " using default empty plugins info list.".format(
                        self.jenkins.server
                    )
                )
                plugins_list = [{"shortName": "", "version": "", "longName": ""}]
            else:
                raise
        logger.debug("Jenkins Plugin Info {0}".format(pformat(plugins_list)))

        return plugins_list

    def get_jobs(self, cache=True):
        if not cache:
            self._jobs = None
            self._job_list = None
        return self.jobs

    def is_managed_job(self, job_name):
        xml = self.jenkins.get_job_config(job_name)
        try:
            out = XML.fromstring(xml.encode("utf-8"))
            description = out.find(".//description").text
            return description.endswith(MAGIC_MANAGE_STRING)
        except (TypeError, AttributeError):
            pass
        return False

    @property
    def plugins_list(self):
        if self._plugins_list is None:
            self._plugins_list = self.get_plugins_info()
        return self._plugins_list

    def delete_old_managed_jobs(self, keep=None):
        jobs = self.get_jobs()
        deleted_jobs = 0
        if keep is None:
            keep = []
        for job in jobs:
            # python-jenkins stores the folder and name as 'fullname'
            # Check if the job was deleted when his parent folder was deleted
            if job["fullname"] not in keep and self.is_job(
                job["fullname"], use_cache=False
            ):
                if self.is_managed_job(job["fullname"]):
                    logger.info(
                        "Removing obsolete jenkins job {0}".format(job["fullname"])
                    )
                    self.delete_job(job["fullname"])
                    deleted_jobs += 1
                else:
                    logger.debug(
                        "Not deleting unmanaged jenkins job %s", job["fullname"]
                    )
            else:
                logger.debug("Keeping job %s", job["fullname"])
        return deleted_jobs

    def delete_jobs(self, jobs):
        if jobs is not None:
            logger.info("Removing jenkins job(s): %s" % ", ".join(jobs))
        for job in jobs:
            self.delete_job(job)
            if self.cache.is_cached(job):
                self.cache.set(job, "")
        self.cache.save()

    def delete_all_jobs(self):
        jobs = self.get_jobs()
        logger.info("Number of jobs to delete:  %d", len(jobs))
        script = (
            "for(job in jenkins.model.Jenkins.theInstance.getAllItems())"
            "       { job.delete(); }"
        )
        self.jenkins.run_script(script)
        # Need to clear the JJB cache after deletion
        self.cache.clear()

    def changed(self, job):
        md5 = job.md5()

        changed = self._jjb_config.builder["ignore_cache"] or self.cache.has_changed(
            job.name, md5
        )
        if not changed:
            logger.debug("'{0}' has not changed".format(job.name))
        return changed

    def exists(self, job):
        exists = self.jenkins.job_exists(job.name)
        if not exists:
            logger.debug("'{0}' does not currently exist".format(job.name))
        return exists

    def update_jobs(
        self,
        xml_jobs,
        output=None,
        n_workers=None,
        existing_only=None,
        config_xml=False,
    ):
        orig = time.time()

        logger.info("Number of jobs generated:  %d", len(xml_jobs))
        xml_jobs.sort(key=AlphanumSort)

        if output and not hasattr(output, "write") and not os.path.isdir(output):
            logger.debug("Creating directory %s" % output)
            try:
                os.makedirs(output)
            except OSError:
                if not os.path.isdir(output):
                    raise

        if output:
            # ensure only wrapped once
            if hasattr(output, "write"):
                output = utils.wrap_stream(output)

            for job in xml_jobs:
                if hasattr(output, "write"):
                    # `output` is a file-like object
                    logger.info("Job name:  %s", job.name)
                    logger.debug("Writing XML to '{0}'".format(output))
                    try:
                        output.write(job.output())
                    except IOError as exc:
                        if exc.errno == errno.EPIPE:
                            # EPIPE could happen if piping output to something
                            # that doesn't read the whole input (e.g.: the UNIX
                            # `head` command)
                            return
                        raise
                    continue

                output_fn = self._setup_output(output, job.name, config_xml)

                logger.debug("Writing XML to '{0}'".format(output_fn))
                with io.open(output_fn, "w", encoding="utf-8") as f:
                    f.write(job.output().decode("utf-8"))
            return xml_jobs, len(xml_jobs)

        # Filter out the jobs that did not change
        logging.debug("Filtering %d jobs for changed jobs", len(xml_jobs))
        step = time.time()
        jobs = [job for job in xml_jobs if self.changed(job)]
        logging.debug("Filtered for changed jobs in %ss", (time.time() - step))

        if existing_only:
            # Filter out the jobs not already in the cache
            logging.debug("Filtering %d jobs for existing jobs", len(jobs))
            step = time.time()
            jobs = [job for job in jobs if self.exists(job)]
            logging.debug("Filtered for existing jobs in %ss", (time.time() - step))

        if not jobs:
            return [], 0

        # Update the jobs
        logging.debug("Updating jobs")
        step = time.time()
        p_params = [{"job": job} for job in jobs]
        results = self.parallel_update_job(n_workers=n_workers, concurrent=p_params)
        logging.debug("Parsing results")
        # generalize the result parsing, as a concurrent job always returns a
        # list
        if len(p_params) in (1, 0):
            results = [results]
        for result in results:
            if isinstance(result, Exception):
                raise result
            else:
                # update in-memory cache
                j_name, j_md5 = result
                self.cache.set(j_name, j_md5)
        # write cache to disk
        self.cache.save()
        logging.debug("Updated %d jobs in %ss", len(jobs), time.time() - step)
        logging.debug("Total run took %ss", (time.time() - orig))
        return jobs, len(jobs)

    @concurrent
    def parallel_update_job(self, job):
        self.fix_disabled(job)
        self.update_job(job.name, job.output().decode("utf-8"))
        return (job.name, job.md5())

    def fix_disabled(self, job):
        # allow disabled to remain the last setting
        # see modules.general.General
        el = job.xml.find("./disabled")
        if el is not None:
            return
        if not self.is_job(job.name):
            return
        info = self.jenkins.get_job_info(job.name)
        disabled = info.get("disabled", False)
        XML.SubElement(job.xml, "disabled").text = str(disabled).lower()

    ################
    # View related #
    ################

    @property
    def views(self):
        if self._views is None:
            # populate views
            self._views = self.jenkins.get_views()
        return self._views

    @property
    def view_list(self):
        if self._view_list is None:
            self._view_list = set(view["name"] for view in self.views)
        return self._view_list

    def get_views(self, cache=True):
        if not cache:
            self._views = None
            self._view_list = None
        return self.views

    def is_view(self, view_name, use_cache=True):
        if use_cache and view_name in self.view_list:
            return True

        # if not exists, use jenkins
        return self.jenkins.view_exists(view_name)

    def is_managed_view(self, view_name):
        xml = self.jenkins.get_view_config(view_name)
        try:
            out = XML.fromstring(xml.encode("utf-8"))
            description = out.find(".//description").text
            return description.endswith(MAGIC_MANAGE_STRING)
        except (TypeError, AttributeError):
            pass
        return False

    def delete_old_managed_views(self, keep=None):
        view_list = self.get_views()
        deleted_views = 0
        if keep is None:
            keep = []
        for view in view_list:
            if view["name"] not in keep and self.is_view(view["name"], use_cache=False):
                if self.is_managed_view(view["name"]):
                    logger.info(
                        "Removing obsolete jenkins view {0}".format(view["name"])
                    )
                    self.delete_view(view["name"])
                    deleted_views += 1
                else:
                    logger.debug("Not deleting unmanaged jenkins view %s", view["name"])
            else:
                logger.debug("Keeping view %s", view["name"])
        return deleted_views

    def delete_view(self, view_name):
        if self.is_view(view_name):
            logger.info("Deleting jenkins view {}".format(view_name))
            self.jenkins.delete_view(view_name)

    def delete_views(self, views):
        if views is not None:
            logger.info("Removing jenkins view(s): %s" % ", ".join(views))
        for view in views:
            self.delete_view(view)
            if self.cache.is_cached(view):
                self.cache.set(view, "")
        self.cache.save()

    def delete_all_views(self):
        views = self.get_views()
        # Jenkins requires at least one view present. Don't remove the first
        # view as it is likely the default view.
        views.pop(0)
        logger.info("Number of views to delete:  %d", len(views))
        for view in views:
            self.delete_view(view["name"])
        # Need to clear the JJB cache after deletion
        self.cache.clear()

    def update_view(self, view_name, xml):
        if self.is_view(view_name):
            logger.info(
                "Reconfiguring jenkins view {0}".format(self._view_format(view_name))
            )
            self.jenkins.reconfig_view(view_name, xml)
        else:
            logger.info(
                "Creating jenkins view {0}".format(self._view_format(view_name))
            )
            self.jenkins.create_view(view_name, xml)

    def update_views(
        self,
        xml_views,
        output=None,
        n_workers=None,
        existing_only=None,
        config_xml=False,
    ):
        orig = time.time()

        logger.info("Number of views generated:  %d", len(xml_views))
        xml_views.sort(key=AlphanumSort)

        if output:
            # ensure only wrapped once
            if hasattr(output, "write"):
                output = utils.wrap_stream(output)

            for view in xml_views:
                if hasattr(output, "write"):
                    # `output` is a file-like object
                    logger.info("View name:  %s", view.name)
                    logger.debug("Writing XML to '{0}'".format(output))
                    try:
                        output.write(view.output())
                    except IOError as exc:
                        if exc.errno == errno.EPIPE:
                            # EPIPE could happen if piping output to something
                            # that doesn't read the whole input (e.g.: the UNIX
                            # `head` command)
                            return
                        raise
                    continue

                output_fn = self._setup_output(output, view.name, config_xml)

                logger.debug("Writing XML to '{0}'".format(output_fn))
                with io.open(output_fn, "w", encoding="utf-8") as f:
                    f.write(view.output().decode("utf-8"))
            return xml_views, len(xml_views)

        # Filter out the views that did not change
        logging.debug("Filtering %d views for changed views", len(xml_views))
        step = time.time()
        views = [view for view in xml_views if self.changed(view)]
        logging.debug("Filtered for changed views in %ss", (time.time() - step))

        if existing_only:
            # Filter out the jobs not already in the cache
            logging.debug("Filtering %d views for existing jobs", len(views))
            step = time.time()
            views = [view for view in views if self.exists(view)]
            logging.debug("Filtered for existing views in %ss", (time.time() - step))

        if not views:
            return [], 0

        # Update the views
        logging.debug("Updating views")
        step = time.time()
        p_params = [{"view": view} for view in views]
        results = self.parallel_update_view(n_workers=n_workers, concurrent=p_params)
        logging.debug("Parsing results")
        # generalize the result parsing, as a concurrent view always returns a
        # list
        if len(p_params) in (1, 0):
            results = [results]
        for result in results:
            if isinstance(result, Exception):
                raise result
            else:
                # update in-memory cache
                v_name, v_md5 = result
                self.cache.set(v_name, v_md5)
        # write cache to disk
        self.cache.save()
        logging.debug("Updated %d views in %ss", len(views), time.time() - step)
        logging.debug("Total run took %ss", (time.time() - orig))
        return views, len(views)

    @concurrent
    def parallel_update_view(self, view):
        self.update_view(view.name, view.output().decode("utf-8"))
        return (view.name, view.md5())



#!/usr/bin/env python
# Copyright (C) 2018 Sorin Sbarnea
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple

import pytest


JobsScenario = namedtuple("JobsScnenario", "name jobs globs found")

jobs_scenarios = [
    JobsScenario("single", jobs=["job1"], globs=[], found=["job1"]),
    JobsScenario("multiple", jobs=["job1", "job2"], globs=[], found=["job1", "job2"]),
    JobsScenario(
        "multiple_with_folder",
        jobs=["folder1", "folder1/job1", "folder1/job2"],
        globs=[],
        found=["folder1", "folder1/job1", "folder1/job2"],
    ),
    JobsScenario(
        "multiple_with_glob",
        jobs=["job1", "job2", "job3"],
        globs=["job[1-2]"],
        found=["job1", "job2"],
    ),
    JobsScenario(
        "multiple_with_multi_glob",
        jobs=["job1", "job2", "job3", "job4"],
        globs=["job1", "job[24]"],
        found=["job1", "job2", "job4"],
    ),
]


@pytest.mark.parametrize(
    "scenario",
    [pytest.param(s, id=s.name) for s in jobs_scenarios],
)
def test_from_jenkins_tests(
    capsys, mocker, default_config_file, execute_jenkins_jobs, scenario
):
    def get_jobs():
        return [{"fullname": fullname} for fullname in scenario.jobs]

    mocker.patch("jenkins_jobs.builder.JenkinsManager.get_jobs", side_effect=get_jobs)

    args = ["--conf", default_config_file, "list"] + scenario.globs
    execute_jenkins_jobs(args)

    expected_out = "\n".join(scenario.found)
    captured = capsys.readouterr()
    assert captured.out.rstrip() == expected_out


YamlScenario = namedtuple("YamlScnenario", "name globs found")

yaml_scenarios = [
    YamlScenario("all", globs=[], found=["bam001", "bar001", "bar002", "baz001"]),
    YamlScenario(
        "some",
        globs=["*am*", "*002", "bar001"],
        found=["bam001", "bar001", "bar002"],
    ),
]


@pytest.mark.parametrize(
    "scenario",
    [pytest.param(s, id=s.name) for s in yaml_scenarios],
)
def test_from_yaml_tests(
    capsys, fixtures_dir, default_config_file, execute_jenkins_jobs, scenario
):
    path = fixtures_dir / "cmd-002.yaml"

    execute_jenkins_jobs(
        ["--conf", default_config_file, "list", "-p", str(path)] + scenario.globs
    )

    expected_out = "\n".join(scenario.found)
    captured = capsys.readouterr()
    assert captured.out.rstrip() == expected_out



import configparser
import pkg_resources
import xml.etree.ElementTree as XML
from pathlib import Path

import pytest
import yaml

from jenkins.plugins import Plugin
from jenkins_jobs.alphanum import AlphanumSort
from jenkins_jobs.config import JJBConfig
from jenkins_jobs.loader import Loader
from jenkins_jobs.modules import project_externaljob
from jenkins_jobs.modules import project_flow
from jenkins_jobs.modules import project_githuborg
from jenkins_jobs.modules import project_matrix
from jenkins_jobs.modules import project_maven
from jenkins_jobs.modules import project_multibranch
from jenkins_jobs.modules import project_multijob
from jenkins_jobs.registry import ModuleRegistry
from jenkins_jobs.xml_config import XmlJob, XmlJobGenerator, XmlViewGenerator
from jenkins_jobs import utils

from jenkins_jobs.roots import Roots
from jenkins_jobs.loader import load_files


# Avoid writing to ~/.cache/jenkins_jobs.
@pytest.fixture(autouse=True)
def job_cache_mocked(mocker):
    mocker.patch("jenkins_jobs.builder.JobCache", autospec=True)


@pytest.fixture
def config_path(scenario):
    return scenario.config_path


@pytest.fixture
def jjb_config(config_path):
    config = JJBConfig(config_path)
    config.validate()
    return config


@pytest.fixture
def mock_iter_entry_points():
    config = configparser.ConfigParser()
    config.read(Path(__file__).parent / "../setup.cfg")
    groups = {}
    for key in config["entry_points"]:
        groups[key] = list()
        for line in config["entry_points"][key].split("\n"):
            if "" == line.strip():
                continue
            groups[key].append(
                pkg_resources.EntryPoint.parse(line, dist=pkg_resources.Distribution())
            )

    def iter_entry_points(group, name=None):
        return (entry for entry in groups[group] if name is None or name == entry.name)

    return iter_entry_points


@pytest.fixture
def input(scenario, jjb_config):
    loader = Loader.empty(jjb_config)
    return loader.load_path(scenario.in_path)


@pytest.fixture
def plugins_info(scenario):
    if not scenario.plugins_info_path.exists():
        return None
    plugin_dict_list = yaml.safe_load(scenario.plugins_info_path.read_text())
    return [Plugin(**plugin_dict) for plugin_dict in plugin_dict_list]


@pytest.fixture
def registry(mocker, mock_iter_entry_points, jjb_config, plugins_info):
    mocker.patch("pkg_resources.iter_entry_points", side_effect=mock_iter_entry_points)
    return ModuleRegistry(jjb_config, plugins_info)


@pytest.fixture
def project(input, registry):
    type_to_class = {
        "maven": project_maven.Maven,
        "matrix": project_matrix.Matrix,
        "flow": project_flow.Flow,
        "githuborg": project_githuborg.GithubOrganization,
        "multijob": project_multijob.MultiJob,
        "multibranch": project_multibranch.WorkflowMultiBranch,
        "multibranch-defaults": project_multibranch.WorkflowMultiBranchDefaults,
        "externaljob": project_externaljob.ExternalJob,
    }
    try:
        class_name = input["project-type"]
    except KeyError:
        return None
    if class_name == "freestyle":
        return None
    cls = type_to_class[class_name]
    return cls(registry)


@pytest.fixture
def expected_output(scenario):
    if not scenario.out_paths:
        # Do not check output if there are no files for it.
        return None
    return "".join(path.read_text() for path in sorted(scenario.out_paths))


@pytest.fixture
def expected_error(scenario):
    if scenario.error_path.exists():
        return scenario.error_path.read_text().rstrip()
    else:
        return None


# Tests use output files directories as expected folder name.
def check_folders(scenario, job_xml_list):
    root_dir = scenario.in_path.parent

    def name_parent(name):
        *dirs, name = name.split("/")
        return "/".join(dirs)

    def path_parent(path):
        if scenario.in_path.is_dir():
            # In directory tests, output file directory does not
            # indicate expected job folder.
            base_dir = scenario.in_path
        else:
            base_dir = root_dir
        dir = str(path.relative_to(base_dir).parent)
        if dir == ".":
            return ""
        else:
            return dir

    actual_dirs = list(sorted(set(name_parent(jx.name) for jx in job_xml_list)))
    expected_dirs = list(sorted(path_parent(path) for path in scenario.out_paths))
    assert expected_dirs == actual_dirs


@pytest.fixture
def check_generator(scenario, input, expected_output, jjb_config, registry, project):
    def check(Generator):
        if project:
            xml = project.root_xml(input)
        else:
            xml = XML.Element("project")

        generator = Generator(registry)
        generator.gen_xml(xml, input)
        pretty_xml = (
            XmlJob(xml, "fixturejob")
            .output()
            .decode()
            .replace("&quot;", '"')  # Ensure compatibility with Python < 3.13
        )
        assert expected_output == pretty_xml

    return check


@pytest.fixture
def check_parser(jjb_config, registry):
    def check(in_path):
        roots = Roots(jjb_config)
        load_files(jjb_config, roots, [in_path])
        registry.set_macros(roots.macros)
        job_data_list = roots.generate_jobs()
        view_data_list = roots.generate_views()
        generator = XmlJobGenerator(registry)
        _ = generator.generateXML(job_data_list)
        _ = generator.generateXML(view_data_list)

    return check


@pytest.fixture
def check_job(scenario, expected_output, jjb_config, registry):
    def check():
        roots = Roots(jjb_config)
        if jjb_config.recursive:
            path_list = [Path(p) for p in utils.recurse_path(str(scenario.in_path))]
        else:
            path_list = [scenario.in_path]
        load_files(jjb_config, roots, path_list)
        registry.set_macros(roots.macros)
        job_data_list = roots.generate_jobs()
        registry.amend_job_dicts(job_data_list)
        generator = XmlJobGenerator(registry)
        job_xml_list = generator.generateXML(job_data_list)
        job_xml_list.sort(key=AlphanumSort)

        pretty_xml = (
            "\n".join(job.output().decode() for job in job_xml_list)
            .strip()
            .replace("\n\n", "\n")
            .replace("&quot;", '"')  # Ensure compatibility with Python < 3.13
        )
        if expected_output is None:
            return
        stripped_expected_output = (
            expected_output.strip().replace("<BLANKLINE>", "").replace("\n\n", "\n")
        )
        assert stripped_expected_output == pretty_xml

        check_folders(scenario, job_xml_list)

    return check


@pytest.fixture
def check_view(scenario, expected_output, jjb_config, registry):
    def check():
        roots = Roots(jjb_config)
        load_files(jjb_config, roots, [scenario.in_path])
        registry.set_macros(roots.macros)
        view_data_list = roots.generate_views()
        generator = XmlViewGenerator(registry)
        view_xml_list = generator.generateXML(view_data_list)
        view_xml_list.sort(key=AlphanumSort)

        pretty_xml = (
            "\n".join(view.output().decode() for view in view_xml_list)
            .strip()
            .replace("\n\n", "\n")
        )
        if expected_output is None:
            return
        stripped_expected_output = (
            expected_output.strip().replace("<BLANKLINE>", "").replace("\n\n", "\n")
        )
        assert stripped_expected_output == pretty_xml

    return check



#
# Copyright (c) 2016 Kien Ha <kienha9922@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import xml.etree.ElementTree as XML
import yaml

import pytest

from jenkins_jobs.errors import InvalidAttributeError
from jenkins_jobs.errors import MissingAttributeError
from jenkins_jobs.errors import JenkinsJobsException
from jenkins_jobs.modules.helpers import (
    convert_mapping_to_xml,
    check_mutual_exclusive_data_args,
)


def test_convert_mapping_to_xml():
    """
    Tests the test_convert_mapping_to_xml_fail_required function
    """

    # Test default values
    default_root = XML.Element("testdefault")
    default_data = yaml.safe_load("string: hello")
    default_mappings = [("default-string", "defaultString", "default")]

    convert_mapping_to_xml(
        default_root, default_data, default_mappings, fail_required=True
    )
    result = default_root.find("defaultString").text
    result == "default"

    # Test user input
    user_input_root = XML.Element("testUserInput")
    user_input_data = yaml.safe_load("user-input-string: hello")
    user_input_mappings = [("user-input-string", "userInputString", "user-input")]

    convert_mapping_to_xml(
        user_input_root, user_input_data, user_input_mappings, fail_required=True
    )
    result = user_input_root.find("userInputString").text
    result == "hello"

    # Test missing required input
    required_root = XML.Element("testrequired")
    required_data = yaml.safe_load("string: hello")
    required_mappings = [("required-string", "requiredString", None)]

    with pytest.raises(MissingAttributeError):
        convert_mapping_to_xml(
            required_root,
            required_data,
            required_mappings,
            fail_required=True,
        )

    # Test invalid user input for list
    user_input_root = XML.Element("testUserInput")
    user_input_data = yaml.safe_load("user-input-string: bye")
    valid_inputs = ["hello"]
    user_input_mappings = [
        ("user-input-string", "userInputString", "user-input", valid_inputs)
    ]

    with pytest.raises(InvalidAttributeError):
        convert_mapping_to_xml(
            user_input_root,
            user_input_data,
            user_input_mappings,
        )

    # Test invalid user input for dict
    user_input_root = XML.Element("testUserInput")
    user_input_data = yaml.safe_load("user-input-string: later")
    valid_inputs = {"hello": "world"}
    user_input_mappings = [
        ("user-input-string", "userInputString", "user-input", valid_inputs)
    ]

    with pytest.raises(InvalidAttributeError):
        convert_mapping_to_xml(
            user_input_root,
            user_input_data,
            user_input_mappings,
        )

    # Test invalid key for dict
    user_input_root = XML.Element("testUserInput")
    user_input_data = yaml.safe_load("user-input-string: world")
    valid_inputs = {"hello": "world"}
    user_input_mappings = [
        ("user-input-string", "userInputString", "user-input", valid_inputs)
    ]

    with pytest.raises(InvalidAttributeError):
        convert_mapping_to_xml(
            user_input_root,
            user_input_data,
            user_input_mappings,
        )


def test_check_mutual_exclusive_data_args_no_mutual_exclusive():
    @check_mutual_exclusive_data_args(0, "foo", "bar")
    @check_mutual_exclusive_data_args(0, "foo", "baz")
    def func(data):
        pass

    func({"baz": "qaz", "bar": "qaz"})


def test_check_mutual_exclusive_data_args_mutual_exclusive():
    @check_mutual_exclusive_data_args(0, "foo", "bar")
    @check_mutual_exclusive_data_args(0, "foo", "baz")
    def func(data):
        pass

    with pytest.raises(JenkinsJobsException):
        func({"foo": "qaz", "bar": "qaz"})



# -*- coding: utf-8 -*-
# Copyright (C) 2021 The Linux Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""
The Github Organization project module handles creating Jenkins Github
Organization jobs, which are made up of multibranch pipelines for all
repositories containing the specified Jenkinsfile(s).
You may specify ``githuborg`` in the ``project-type`` attribute of
the :ref:`Job` definition.

Plugins required:

* :jenkins-plugins:`GitHub Branch Source Plugin <github-branch-source>`

:Job Parameters:

    * **github-org** (`dict`): Refer to
      :func:`github_org <github_org>` for documentation.

    * **periodic-folder-trigger** (`str`): How often to scan for new branches
      or pull/change requests. Valid values: 1m, 2m, 5m, 10m, 15m, 20m, 25m,
      30m, 1h, 2h, 4h, 8h, 12h, 1d, 2d, 1w, 2w, 4w. (default none)
    * **prune-dead-branches** (`bool`): If dead branches upon check should
      result in their job being dropped. (default true)
    * **number-to-keep** (`int`): How many builds should be kept.
      (default '-1, all')
    * **days-to-keep** (`int`): For how many days should a build be kept.
      (default '-1, forever')
    * **script-path** (`str`): Path to Jenkinsfile, relative to workspace.
      (default 'Jenkinsfile')

Job examples:

.. literalinclude:: /../../tests/githuborg/fixtures/minimal.yaml

.. literalinclude:: /../../tests/githuborg/fixtures/githuborg-job-full.yaml

"""
import collections
import logging
import xml.etree.ElementTree as XML
import jenkins_jobs.modules.base
import jenkins_jobs.modules.helpers as helpers
import jenkins_jobs.modules.project_multibranch as multibranch

from jenkins_jobs.modules.scm import git_extensions
from jenkins_jobs.errors import InvalidAttributeError

logger = logging.getLogger(str(__name__))


class GithubOrganization(jenkins_jobs.modules.base.Base):
    sequence = 0
    jenkins_class = "jenkins.branch.OrganizationFolder"

    def root_xml(self, data):
        xml_parent = XML.Element(self.jenkins_class)
        xml_parent.attrib["plugin"] = "branch-api"
        XML.SubElement(xml_parent, "properties")

        ################
        # Folder Views #
        ################

        folderViews = XML.SubElement(
            xml_parent,
            "folderViews",
            {"class": "jenkins.branch.OrganizationFolderViewHolder"},
        )

        XML.SubElement(folderViews, "owner", {"reference": "../.."})

        ##################
        # Health Metrics #
        ##################

        hm = XML.SubElement(xml_parent, "healthMetrics")
        hm_path = "com.cloudbees.hudson.plugins.folder.health" ".WorstChildHealthMetric"
        hm_plugin = XML.SubElement(hm, hm_path, {"plugin": "cloudbees-folder"})
        XML.SubElement(hm_plugin, "nonRecursive").text = "false"

        ########
        # Icon #
        ########

        icon = XML.SubElement(
            xml_parent, "icon", {"class": "jenkins.branch.MetadataActionFolderIcon"}
        )
        XML.SubElement(
            icon, "owner", {"class": self.jenkins_class, "reference": "../.."}
        )

        ########################
        # Orphan Item Strategy #
        ########################

        ois_default_strategy = (
            "com.cloudbees.hudson.plugins."
            "folder.computed.DefaultOrphanedItemStrategy"
        )
        ois = XML.SubElement(
            xml_parent,
            "orphanedItemStrategy",
            {"class": ois_default_strategy, "plugin": "cloudbees-folder"},
        )

        ois_mapping = [
            ("prune-dead-branches", "pruneDeadBranches", True, [True, False]),
            ("days-to-keep", "daysToKeep", -1),
            ("number-to-keep", "numToKeep", -1),
        ]
        helpers.convert_mapping_to_xml(ois, data, ois_mapping)

        ###########################
        # Periodic Folder Trigger #
        ###########################

        triggers = XML.SubElement(xml_parent, "triggers")

        # Valid options for the periodic trigger interval.
        pft_map = collections.OrderedDict(
            [
                ("1m", ("* * * * *", "60000")),
                ("2m", ("*/2 * * * *", "120000")),
                ("5m", ("*/5 * * * *", "300000")),
                ("10m", ("H/6 * * * *", "600000")),
                ("15m", ("H/6 * * * *", "900000")),
                ("20m", ("H/3 * * * *", "1200000")),
                ("25m", ("H/3 * * * *", "1500000")),
                ("30m", ("H/2 * * * *", "1800000")),
                ("1h", ("H * * * *", "3600000")),
                ("2h", ("H * * * *", "7200000")),
                ("4h", ("H * * * *", "14400000")),
                ("8h", ("H * * * *", "28800000")),
                ("12h", ("H H * * *", "43200000")),
                ("1d", ("H H * * *", "86400000")),
                ("2d", ("H H * * *", "172800000")),
                ("1w", ("H H * * *", "604800000")),
                ("2w", ("H H * * *", "1209600000")),
                ("4w", ("H H * * *", "2419200000")),
            ]
        )

        pft_val = data.get("periodic-folder-trigger")
        if pft_val:
            if not pft_map.get(pft_val):
                raise InvalidAttributeError(
                    "periodic-folder-trigger", pft_val, pft_map.keys()
                )

            pft_path = (
                "com.cloudbees.hudson.plugins.folder.computed." "PeriodicFolderTrigger"
            )
            pft = XML.SubElement(triggers, pft_path, {"plugin": "cloudbees-folder"})
            XML.SubElement(pft, "spec").text = pft_map[pft_val][0]
            XML.SubElement(pft, "interval").text = pft_map[pft_val][1]

        ##############
        # Navigators #
        ##############

        navigators = XML.SubElement(xml_parent, "navigators")
        navigators_plugin = XML.SubElement(
            navigators,
            "org.jenkinsci.plugins.github__branch__source.GitHubSCMNavigator",
            {"plugin": "github-branch-source"},
        )
        github_org(navigators_plugin, data.get("github-org"))

        ###########
        # Factory #
        ###########

        fopts_map = [("script-path", "scriptPath", "Jenkinsfile")]

        project_factories = XML.SubElement(xml_parent, "projectFactories")
        factory = XML.SubElement(
            project_factories,
            "org.jenkinsci.plugins.workflow.multibranch.WorkflowMultiBranchProjectFactory",
            {"plugin": "workflow-multibranch"},
        )
        helpers.convert_mapping_to_xml(factory, data, fopts_map, fail_required=False)

        ####################
        # Build Strategies #
        ####################

        if data.get("github-org").get("build-strategies", None):
            multibranch.build_strategies(xml_parent, data.get("github-org"))

        return xml_parent


def github_org(xml_parent, data):
    r"""Configure GitHub Organization and SCM settings.

    :arg str repo-owner: Specify the name of the GitHub Organization or
        GitHub User Account. (required)
    :arg str api-uri: The GitHub API uri for hosted / on-site GitHub. Must
        first be configured in Global Configuration. (default GitHub)
    :arg str branch-discovery: Discovers branches on the repository.
        Valid options: no-pr, only-pr, all, false. (default 'no-pr')
    :arg str repo-name-regex: Regular expression used to match repository names
        within the organization. (optional)
        Requires the :jenkins-plugins:`SCM API plugin <scm-api>`.
    :arg list build-strategies: Provides control over whether to build a branch
        (or branch like things such as change requests and tags) whenever it is
        discovered initially or a change from the previous revision has been
        detected. (optional)
        Refer to :func:`~build_strategies <project_multibranch.build_strategies>`.
    :arg str credentials-id: Credentials used to scan branches and pull
        requests, check out sources and mark commit statuses. (optional)
    :arg str discover-pr-forks-strategy: Fork strategy. Valid options:
        merge-current, current, both, false. (default 'merge-current')
    :arg str discover-pr-forks-trust: Discovers pull requests where the origin
        repository is a fork of the target repository.
        Valid options: contributors, everyone, permission or nobody.
        (default 'contributors')
    :arg str discover-pr-origin: Discovers pull requests where the origin
        repository is the same as the target repository.
        Valid options: merge-current, current, both, false.  (default 'merge-current')
    :arg bool discover-tags: Discovers tags on the repository.
        (default false)
    :arg list head-pr-filter-behaviors: Definition of Filter Branch PR behaviors.
        Requires the :jenkins-plugins:`SCM Filter Branch PR Plugin
        <scm-filter-branch-pr>`. Refer to
        :func:`~add_filter_branch_pr_behaviors <project_multibranch.add_filter_branch_pr_behaviors>`.
    :arg dict notification-context: Change the default GitHub check notification
        context from "continuous-integration/jenkins/SUFFIX" to a custom label / suffix.
        (set a label and suffix to true or false, optional)
        Requires the :jenkins-plugins:`Github Custom Notification Context SCM
        Behaviour <github-scm-trait-notification-context>`.
        Refer to :func:`~add_notification_context_trait <project_multibranch.add_notification_context_trait>`.
    :arg dict property-strategies: Provides control over how to build a branch
        (like to disable SCM triggering or to override the pipeline durability)
        (optional)
        Refer to :func:`~property_strategies <project_multibranch.property_strategies>`.
    :arg bool exclude-archived-repositories: Whether archived repositories are
        excluded when scanning an organization. (default: false) (optional)
    :arg bool ssh-checkout: Checkout over SSH.

        * **credentials** ('str'): Credentials to use for
            checkout of the repo over ssh.

    :extensions:

        * **clean** (`dict`)
            * **after** (`dict`) - Clean the workspace after checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
            * **before** (`dict`) - Clean the workspace before checkout
                * **remove-stale-nested-repos** (`bool`) - Deletes untracked
                  submodules and any other subdirectories which contain .git directories
                  (default false)
        * **depth** (`int`) - Set shallow clone depth (default 1)
        * **disable-pr-notifications** (`bool`) - Disable default github status
            notifications on pull requests (default false) (Requires the
            :jenkins-plugins:`GitHub Branch Source Plugin
            <disable-github-multibranch-status>`.)
        * **do-not-fetch-tags** (`bool`) - Perform a clone without tags
            (default false)
        * **lfs-pull** (`bool`) - Call git lfs pull after checkout
            (default false)
        * **prune** (`bool`) - Prune remote branches (default false)
        * **refspecs** (`list(str)`): Which refspecs to fetch.
        * **shallow-clone** (`bool`) - Perform shallow clone (default false)
        * **sparse-checkout** (dict)
            * **paths** (list) - List of paths to sparse checkout. (optional)
        * **submodule** (`dict`)
            * **disable** (`bool`) - By disabling support for submodules you
              can still keep using basic git plugin functionality and just have
              Jenkins to ignore submodules completely as if they didn't exist.
            * **recursive** (`bool`) - Retrieve all submodules recursively
              (uses '--recursive' option which requires git>=1.6.5)
            * **tracking** (`bool`) - Retrieve the tip of the configured
              branch in .gitmodules (Uses '\-\-remote' option which requires
              git>=1.8.2)
            * **parent-credentials** (`bool`) - Use credentials from default
              remote of parent repository (default false).
            * **reference-repo** (`str`) - Path of the reference repo to use
              during clone (optional)
            * **timeout** (`int`) - Specify a timeout (in minutes) for
              submodules operations (default 10).
        * **timeout** (`str`) - Timeout for git commands in minutes (optional)
        * **use-author** (`bool`): Use author rather than committer in Jenkin's
            build changeset (default false)
        * **wipe-workspace** (`bool`) - Wipe out workspace before build
            (default true)

    Job examples:

    .. literalinclude:: /../../tests/githuborg/fixtures/minimal.yaml

    .. literalinclude:: /../../tests/githuborg/fixtures/github-org-full.yaml

    """
    github_path = "org.jenkinsci.plugins.github_branch_source"
    github_path_dscore = "org.jenkinsci.plugins.github__branch__source"

    mapping = [("repo-owner", "repoOwner", None)]
    helpers.convert_mapping_to_xml(xml_parent, data, mapping, fail_required=True)

    mapping_optional = [
        ("api-uri", "apiUri", None),
        ("credentials-id", "credentialsId", None),
    ]
    helpers.convert_mapping_to_xml(
        xml_parent, data, mapping_optional, fail_required=False
    )

    traits = XML.SubElement(xml_parent, "traits")

    # no-pr value is assumed if branch-discovery not mentioned.
    if data.get("branch-discovery", "no-pr"):
        bd = XML.SubElement(
            traits, "".join([github_path_dscore, ".BranchDiscoveryTrait"])
        )
        bd_strategy = {"no-pr": "1", "only-pr": "2", "all": "3"}
        bd_mapping = [("branch-discovery", "strategyId", "no-pr", bd_strategy)]
        helpers.convert_mapping_to_xml(bd, data, bd_mapping, fail_required=True)

    if data.get("ssh-checkout", None):
        cossh = XML.SubElement(
            traits, "".join([github_path_dscore, ".SSHCheckoutTrait"])
        )
        if not isinstance(data.get("ssh-checkout"), bool):
            cossh_credentials = [("credentials", "credentialsId", "")]
            helpers.convert_mapping_to_xml(
                cossh, data.get("ssh-checkout"), cossh_credentials, fail_required=True
            )

    if data.get("discover-tags", False):
        XML.SubElement(traits, "".join([github_path_dscore, ".TagDiscoveryTrait"]))

    if data.get("discover-pr-forks-strategy", "merged-current"):
        dprf = XML.SubElement(
            traits, "".join([github_path_dscore, ".ForkPullRequestDiscoveryTrait"])
        )
        dprf_strategy = {"merge-current": "1", "current": "2", "both": "3"}
        dprf_mapping = [
            ("discover-pr-forks-strategy", "strategyId", "merge-current", dprf_strategy)
        ]
        helpers.convert_mapping_to_xml(dprf, data, dprf_mapping, fail_required=True)

        trust = data.get("discover-pr-forks-trust", "contributors")
        trust_map = {
            "contributors": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustContributors"]
            ),
            "everyone": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustEveryone"]
            ),
            "permission": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustPermission"]
            ),
            "nobody": "".join(
                [github_path, ".ForkPullRequestDiscoveryTrait$TrustNobody"]
            ),
        }
        if trust not in trust_map:
            raise InvalidAttributeError(
                "discover-pr-forks-trust", trust, trust_map.keys()
            )
        XML.SubElement(dprf, "trust").attrib["class"] = trust_map[trust]

    dpro_strategy = data.get("discover-pr-origin", "merge-current")
    if dpro_strategy:
        dpro = XML.SubElement(
            traits, "".join([github_path_dscore, ".OriginPullRequestDiscoveryTrait"])
        )
        dpro_strategy_map = {"merge-current": "1", "current": "2", "both": "3"}
        if dpro_strategy not in dpro_strategy_map:
            raise InvalidAttributeError(
                "discover-pr-origin", dpro_strategy, dpro_strategy_map.keys()
            )
        dpro_mapping = [
            ("discover-pr-origin", "strategyId", "merge-current", dpro_strategy_map)
        ]
        helpers.convert_mapping_to_xml(dpro, data, dpro_mapping, fail_required=True)

    if data.get("head-filter-regex", None):
        rshf = XML.SubElement(traits, "jenkins.scm.impl.trait.RegexSCMHeadFilterTrait")
        XML.SubElement(rshf, "regex").text = data.get("head-filter-regex")

    if data.get("repo-name-regex", None):
        rssf = XML.SubElement(
            traits,
            "jenkins.scm.impl.trait.RegexSCMSourceFilterTrait",
            {"plugin": "scm-api"},
        )
        XML.SubElement(rssf, "regex").text = data.get("repo-name-regex")

    if data.get("head-pr-filter-behaviors", None):
        multibranch.add_filter_branch_pr_behaviors(
            traits, data.get("head-pr-filter-behaviors")
        )

    if data.get("property-strategies", None):
        multibranch.property_strategies(xml_parent, data)

    multibranch.add_notification_context_trait(traits, data)

    # handle the default git extensions like:
    # - clean
    # - shallow-clone
    # - timeout
    # - do-not-fetch-tags
    # - submodule
    # - prune
    # - wipe-workspace
    # - use-author
    # - lfs-pull
    git_extensions(traits, data)

    if data.get("refspecs"):
        refspec_trait = XML.SubElement(
            traits,
            "jenkins.plugins.git.traits.RefSpecsSCMSourceTrait",
            {"plugin": "git"},
        )
        templates = XML.SubElement(refspec_trait, "templates")
        refspecs = data.get("refspecs")
        for refspec in refspecs:
            e = XML.SubElement(
                templates,
                (
                    "jenkins.plugins.git.traits"
                    ".RefSpecsSCMSourceTrait_-RefSpecTemplate"
                ),
            )
            XML.SubElement(e, "value").text = refspec

    # github-only extensions
    disable_github_status_path_dscore = (
        "com.adobe.jenkins.disable__github__multibranch__status"
    )
    if data.get("disable-pr-notifications", False):
        XML.SubElement(
            traits,
            "".join([disable_github_status_path_dscore, ".DisableStatusUpdateTrait"]),
            {"plugin": "disable-github-multibranch-status"},
        )

    if data.get("exclude-archived-repositories", False):
        XML.SubElement(
            traits,
            "".join([github_path_dscore, ".ExcludeArchivedRepositoriesTrait"]),
        )



from operator import attrgetter
from pathlib import Path

import pytest

from tests.enum_scenarios import scenario_list


fixtures_dir = Path(__file__).parent / "view_fixtures"


@pytest.fixture(
    params=scenario_list(fixtures_dir),
    ids=attrgetter("name"),
)
def scenario(request):
    return request.param


def test_yaml_snippet(check_view):
    check_view()
